from UQpy.Distributions import *
import numpy as np
import scipy.stats as stats

class Nataf:
    """
    Transform random variables using the Nataf or Inverse Nataf transformation

    **Inputs:**

    * **dist_object** ((list of ) ``Distribution`` object(s)):
        Probability distribution of each random variable. Must be an object of type
        ``DistributionContinuous1D`` or ``JointInd``.

    * **corr_x** (`ndarray`):
        The correlation  matrix (:math:`\mathbf{C_X}`) of the random vector **X** .

        Default: The identity matrix.

        If ``corr_x`` is specified, the ``Nataf`` class invokes the ITAM to compute ``corr_z``.

    * **corr_z** (`ndarray`):
        The correlation  matrix (:math:`\mathbf{C_Z}`) of the standard normal random vector **Z** .

        Default: The identity matrix.

        If ``corr_z`` is specified, the ``Nataf`` class computes the correlation distortion integral above to solve for
        ``corr_x``.

    * **itam_threshold1** (`float`):
        If ``corr_x`` is specified, this specifies the threshold value for the error in the `ITAM` method (see
        ``Utilities`` module) given by:

        .. math:: \epsilon_1 = ||\mathbf{C_X}^{target} - \mathbf{C_X}^{computed}||

        Default: 0.001

    * **itam_threshold2** (`float`):
        If ``corr_x`` is specified, this specifies the threshold value for the error difference between iterations in
        the `ITAM` method (see ``Utilities`` module) given by:

        .. math:: \epsilon_1^{i} - \epsilon_1^{i-1}

        for iteration :math:`i`.

        Default: 0.01

    * **beta** (`float`):
        A parameter selected to optimize convergence speed and desired accuracy of the ITAM method.

        Default: 1.0

    * **samples_x** (`ndarray`):
            Random vector of shape ``(nsamples, dimension)`` with prescribed probability distributions.

            If `samples_x` is provided, the ``Nataf`` class transforms them to `samples_z`.

    * **samples_z** (`ndarray`):
            Standard normal random vector of shape ``(nsamples, dimension)``

            If `samples_z` is provided, the ``Nataf`` class transforms them to `samples_x`.

    * **jacobian** ('Boolean'):
            A boolean whether to return the jacobian of the transformation.

            Default: False

    * **itam_max_iter** (`int`):
        Maximum number of iterations for the ITAM method.

        Default: 100

    * **verbose** (`Boolean`):
        A boolean declaring whether to write text to the terminal.

        Default: ``False``

    **Attributes:**

    * **corr_z** (`ndarray`):
        Distorted correlation matrix (:math:`\mathbf{C_Z}`) of the standard normal vector **Z**.

    * **corr_x** (`ndarray`):
        Distorted correlation matrix (:math:`\mathbf{C_X}`) of the random vector **X**.

    * **H** (`ndarray`):
        The lower triangular matrix resulting from the Cholesky decomposition of the correlation matrix
        :math:`\mathbf{C_Z}`.

    * **itam_error1** (`list`)
        List of ITAM errors for each iteration

    * **itam_error2** (`list`)
        List of ITAM difference errors for each iteration

    * **samples_x** (`ndarray`):
        Random vector of shape ``(nsamples, dimension)`` with prescribed probability distributions.

    * **samples_z** (`ndarray`):
        Standard normal random vector of shape ``(nsamples, dimension)``

    * **jxz** (`ndarray`):
        The Jacobian of the transformation of shape ``(dimension, dimension)``.

    * **jzx** (`ndarray`):
        The Jacobian of the transformation of shape ``(dimension, dimension)``.


    **Methods:**
    """

    def __init__(self, dist_object, samples_x=None, samples_z=None, jacobian=False, corr_z=None, corr_x=None, beta=None,
                 itam_threshold1=None, itam_threshold2=None, itam_max_iter=None, verbose=False):

        if isinstance(dist_object, list):
            self.dimension = len(dist_object)
            for i in range(len(dist_object)):
                if not isinstance(dist_object[i], (DistributionContinuous1D, JointInd)):
                    raise TypeError('UQpy: A  ``DistributionContinuous1D`` or ``JointInd`` object '
                                    'must be provided.')
        else:
            if not isinstance(dist_object, (DistributionContinuous1D, JointInd)):
                raise TypeError('UQpy: A  ``DistributionContinuous1D``  or ``JointInd`` object must be provided.')

        self.dist_object = dist_object
        self.corr_x = corr_x
        self.corr_z = corr_z
        self.samples_x = samples_x
        self.samples_z = samples_z
        self.jacobian = jacobian
        self.verbose = verbose
        self.itam_max_iter = itam_max_iter
        self.jzx = None
        self.jxz = None

        self.beta = beta
        self.itam_threshold1 = itam_threshold1
        self.itam_threshold2 = itam_threshold2
        self.corr_x = corr_x
        self.dist_object = dist_object

        if corr_x is None and corr_z is None:
            self.corr_x = np.eye(self.dimension)
            self.corr_z = np.eye(self.dimension)
        elif corr_x is not None:
            if np.all(np.equal(self.corr_x, np.eye(self.dimension))):
                self.corr_z = self.corr_x
            elif all(isinstance(x, Normal) for x in dist_object):
                self.corr_z = self.corr_x
            else:
                self.corr_z, self.itam_error1, self.itam_error2 = self.itam(self.dist_object, self.corr_x, self.beta,
                                                                            self.itam_threshold1, self.itam_threshold2,
                                                                            self.verbose)
        elif corr_z is not None:
            if np.all(np.equal(self.corr_z, np.eye(self.dimension))):
                self.corr_x = self.corr_z
            elif all(isinstance(x, Normal) for x in dist_object):
                self.corr_x = self.corr_z
            else:
                self.corr_x = self.distortion_z2x(self.dist_object, self.corr_z)

        from scipy.linalg import cholesky
        self.H = cholesky(self.corr_z, lower=True)

        if self.samples_x is not None or self.samples_z is not None:
            self.run(self.samples_x, self.samples_z, self.jacobian)

    def run(self, samples_x=None, samples_z=None, jacobian=False):
        """
        Execute the Nataf transformation or its inverse.

        If `samples_x` is provided, the ``run`` method performs the Nataf transformation. If `samples_z` is provided, \
        the ``run`` method performs the inverse Nataf transformation.

        ** Input:**

        * **samples_x** or **samples_z** (`ndarray`):
            Random vector **X**  with prescribed probability distributions or standard normal random vector **Z** of
            shape``(nsamples, dimension)``.

        * **jacobian** (`Boolean`):
            The jacobian of the transformation of shape ``(dimension, dimension)``.

            Default: ``False``

        """
        self.jacobian = jacobian

        if samples_x is not None:
            self.samples_x = samples_x
            if jacobian is False:
                self.samples_z = self._transform_x2z(self.samples_x)
            elif jacobian is True:
                self.samples_z, self.jxz = self._transform_x2z(self.samples_x, jacobian=self.jacobian)

        if samples_z is not None:
            self.samples_z = samples_z
            if self.jacobian is False:
                self.samples_x = self._transform_z2x(self.samples_z)
            elif self.jacobian is True:
                self.samples_x, self.jzx = self._transform_z2x(self.samples_z, jacobian=self.jacobian)

    @staticmethod
    def itam(dist_object, corr_x,  itam_max_iter=None, beta=None, itam_threshold1=None, itam_threshold2=None,
             verbose=None):
        """
        Calculate the correlation matrix :math:`\mathbf{C_Z}` of the standard normal random vector
        :math:`\mathbf{Z}` given the correlation matrix :math:`\mathbf{C_X}` of the random vector :math:`\mathbf{X}`
        using the `ITAM` method [3]_.

        **Inputs:**

        * **dist_object** ((list of ) ``Distribution`` object(s)):
            Probability distribution of each random variable. Must be an object of type
            ``DistributionContinuous1D`` or ``JointInd``.

            `dist_object` must have a ``cdf`` method.

        * **corr_x** (`ndarray`):
            The correlation  matrix (:math:`\mathbf{C_X}`) of the random vector **X** .

            Default: The ``identity`` matrix.

        * **itam_threshold1** (`float`):
            If ``corr_x`` is specified, this specifies the threshold value for the error in the `ITAM` method (see
            ``Utilities`` module) given by:

            .. math:: \epsilon_1 = ||\mathbf{C_X}^{target} - \mathbf{C_X}^{computed}||
        * **itam_max_iter** (`int`):
            Maximum number of iterations for the ITAM method.

            Default: 100

        * **itam_threshold1** (`float`):
            A threshold value for the relative difference between the non-Gaussian correlation function and the
            underlying Gaussian.

            Default: 0.001

        * **itam_threshold2** (`float`):
            If ``corr_x`` is specified, this specifies the threshold value for the error difference between iterations
            in the `ITAM` method (see ``Utilities`` module) given by:

            .. math:: \epsilon_1^{i} - \epsilon_1^{i-1}

            for iteration :math:`i`.

            Default: 0.01

        * **beta** (`float`):
            A parameters selected to optimize convergence speed and desired accuracy of the ITAM method (see [2]_).

            Default: 1.0

        * **verbose** (`Boolean`):
            A boolean declaring whether to write text to the terminal.

            Default: False

        **Output/Returns:**

        * **corr_z** (`ndarray`):
            Distorted correlation matrix (:math:`\mathbf{C_Z}`) of the standard normal vector **Z**.

        * **itam_error1** (`list`)
            List of ITAM errors for each iteration

        * **itam_error2** (`list`)
            List of ITAM difference errors for each iteration

        """

        if itam_max_iter is None:
            itam_max_iter = 100
        if beta is None:
            beta = 1.0
        if itam_threshold1 is None:
            itam_threshold1 = 0.001
        if itam_threshold2 is None:
            itam_threshold2 = 0.1
        if verbose is None:
            verbose = False

        # Initial Guess
        corr_z0 = corr_x
        corr_z = np.zeros_like(corr_z0)
        # Iteration Condition
        itam_error1 = list()
        itam_error2 = list()
        itam_error1.append(100.0)
        itam_error2.append(abs(itam_error1[0] - 0.1) / 0.1)

        if verbose:
            print("UQpy: Initializing Iterative Translation Approximation Method (ITAM)")

        for k in range(itam_max_iter):
            error0 = itam_error1[k]
            from UQpy.Utilities import nearest_psd
            corr0 = Nataf.distortion_z2x(dist_object, corr_z0, verbose)

            max_ratio = np.amax(np.ones((len(corr_x), len(corr_x))) / abs(corr_z0))

            corr_z = np.nan_to_num((corr_x / corr0) ** beta * corr_z0)

            # Do not allow off-diagonal correlations to equal or exceed one
            corr_z[corr_z < -1.0] = (max_ratio + 1) / 2 * corr_z0[corr_z < -1.0]
            corr_z[corr_z > 1.0] = (max_ratio + 1) / 2 * corr_z0[corr_z > 1.0]

            corr_z = np.array(nearest_psd(corr_z))

            corr_z0 = corr_z.copy()

            itam_error1.append(np.linalg.norm(corr_x - corr0))
            itam_error2.append(abs(itam_error1[-1] - error0) / error0)

            if verbose:
                print("UQpy: ITAM iteration number ", k)
                print("UQpy: Current error, ", itam_error1[-1], itam_error2[-1])

            if itam_error1[k] <= itam_threshold1 and itam_error2[k] <= itam_threshold2:
                break

        if verbose:
            print("UQpy: ITAM Done.")

        return corr_z, itam_error1, itam_error2

    @staticmethod
    def distortion_z2x(dist_object, corr_z, verbose=None):
        """
        This is a method to calculate the correlation matrix :math:`\mathbf{C_x}` of the random vector
        :math:`\mathbf{x}`  given the correlation matrix :math:`\mathbf{C_z}` of the standard normal random vector
        :math:`\mathbf{z}`.

        This method is part of the ``Nataf`` class.

        **Inputs:**

        * **dist_object** ((list of ) ``Distribution`` object(s)):
                Probability distribution of each random variable. Must be an object of type
                ``DistributionContinuous1D`` or ``JointInd``.

                **dist_object** must have an ``icdf`` method.

        * **corr_z** (`ndarray`):
            The correlation  matrix (:math:`\mathbf{C_z}`) of the standard normal vector **Z** .

            Default: The ``identity`` matrix.

        * **verbose** (`Boolean`):
            A boolean declaring whether to write text to the terminal.

            Default: ``False``

        **Output/Returns:**

        * **corr_x** (`ndarray`):
            Distorted correlation matrix (:math:`\mathbf{C_X}`) of the random vector **X**.

        """

        if verbose is None:
            verbose = False

        z_max = 8
        z_min = -z_max
        points, weights = np.polynomial.legendre.leggauss(1024)
        points = - (0.5 * (points + 1) * (z_max - z_min) + z_min)
        weights = weights * (0.5 * (z_max - z_min))

        xi = np.tile(points, [1024, 1])
        xi = xi.flatten(order='F')
        eta = np.tile(points, 1024)

        first = np.tile(weights, 1024)
        first = np.reshape(first, [1024, 1024])
        second = np.transpose(first)

        weights2d = first * second
        w2d = weights2d.flatten()

        def bivariate_normal(ksi, psi, rho):
            return (1 / (2 * np.pi * np.sqrt(1 - rho ** 2)) *
                    np.exp(-1 / (2 * (1 - rho ** 2)) *
                    (ksi ** 2 - 2 * rho * ksi * psi + psi ** 2)))

        corr_x = np.ones_like(corr_z)
        if verbose:
            print('UQpy: Computing Nataf correlation distortion...')
        from UQpy.Distributions import JointInd
        if isinstance(dist_object, JointInd):
            if all(hasattr(m, 'moments') for m in dist_object.marginals) and \
                    all(hasattr(m, 'icdf') for m in dist_object.marginals):
                for i in range(len(dist_object.marginals)):
                    i_cdf_i = dist_object.marginals[i].icdf
                    mi = dist_object.marginals[i].moments()
                    if not (np.isfinite(mi[0]) and np.isfinite(mi[1])):
                        raise RuntimeError("UQpy: The marginal distributions need to have finite mean and variance.")
                    for j in range(i + 1, len(dist_object.marginals)):
                        i_cdf_j = dist_object.marginals[j].icdf
                        mj = dist_object.marginals[j].moments()
                        if not (np.isfinite(mj[0]) and np.isfinite(mj[1])):
                            raise RuntimeError(
                                "UQpy: The marginal distributions need to have finite mean and variance.")

                        tmp_f_xi = (i_cdf_j(np.atleast_2d(stats.norm.cdf(xi)).T) - mj[0] ** 2)
                        tmp_f_eta = (i_cdf_i(np.atleast_2d(stats.norm.cdf(eta)).T) - mi[0] ** 2)

                        phi2 = bivariate_normal(xi, eta, corr_z[i, j])

                        corr_x[i, j] = 1/(np.sqrt(mj[1]) * np.sqrt(mi[1])) * np.sum(tmp_f_xi * tmp_f_eta * w2d * phi2)
                        corr_x[j, i] = corr_x[i, j]

        elif isinstance(dist_object, list):
            if all(hasattr(m, 'moments') for m in dist_object) and \
                    all(hasattr(m, 'icdf') for m in dist_object):
                for i in range(len(dist_object)):
                    i_cdf_i = dist_object[i].icdf
                    mi = dist_object[i].moments()
                    if not (np.isfinite(mi[0]) and np.isfinite(mi[1])):
                        raise RuntimeError("UQpy: The marginal distributions need to have finite mean and variance.")

                    for j in range(i + 1, len(dist_object)):
                        i_cdf_j = dist_object[j].icdf
                        mj = dist_object[j].moments()
                        if not (np.isfinite(mj[0]) and np.isfinite(mj[1])):
                            raise RuntimeError(
                                "UQpy: The marginal distributions need to have finite mean and variance.")

                        tmp_f_xi = (i_cdf_j(np.atleast_2d(stats.norm.cdf(xi)).T) - mj[0])
                        tmp_f_eta = (i_cdf_i(np.atleast_2d(stats.norm.cdf(eta)).T) - mi[0])
                        phi2 = bivariate_normal(xi, eta, corr_z[i, j])

                        corr_x[i, j] = 1/(np.sqrt(mj[1]) * np.sqrt(mi[1])) * np.sum(tmp_f_xi * tmp_f_eta * w2d * phi2)
                        corr_x[j, i] = corr_x[i, j]

        if verbose:
            print('UQpy: Done.')
        return corr_x

    def _transform_x2z(self, samples_x, jacobian=False):
        """
        This is a method to transform a vector :math:`\mathbf{x}` of  samples with marginal distributions
        :math:`f_i(x_i)` and cumulative distributions :math:`F_i(x_i)` to a vector :math:`\mathbf{z}` of standard normal
        samples  according to: :math:`Z_{i}=\Phi^{-1}(F_i(X_{i}))`, where :math:`\Phi` is the cumulative
        distribution function of a standard  normal variable.

        This method is part of the ``Nataf`` class.

        **Inputs:**

        * **samples_x** (`ndarray`):
            Random vector of shape ``(nsamples, dimension)`` with prescribed probability distributions.

        * **jacobian** ('Boolean'):
            A boolean whether to return the jacobian of the transformation.

            Default: False

        **Outputs:**

        * **samples_z** (`ndarray`):
            Standard normal random vector of shape ``(nsamples, dimension)``.

        * **jxz** (`ndarray`):
            The jacobian of the transformation of shape ``(dimension, dimension)``.

        """

        m, n = np.shape(samples_x)
        samples_z = None

        if isinstance(self.dist_object, JointInd):
            if all(hasattr(m, 'cdf') for m in self.dist_object.marginals):
                samples_z = np.zeros_like(samples_x)
                for j in range(len(self.dist_object.marginals)):
                    samples_z[:, j] = stats.norm.ppf(self.dist_object.marginals[j].cdf(samples_x[:, j]))
        elif isinstance(self.dist_object, DistributionContinuous1D):
            samples_z = stats.norm.ppf(self.dist_object.cdf(samples_x))
        else:
            samples_z = np.zeros_like(samples_x)
            for j in range(n):
                samples_z[:, j] = stats.norm.ppf(self.dist_object[j].cdf(samples_x[:, j]))

        if not jacobian:
            return samples_z
        else:
            jac = np.zeros(shape=(n, n))
            jxz = [None] * m
            for i in range(m):
                for j in range(n):
                    xi = np.array([samples_x[i, j]])
                    zi = np.array([samples_z[i, j]])
                    jac[j, j] = stats.norm.pdf(zi) / self.dist_object[j].pdf(xi)
                jxz[i] = np.linalg.solve(jac, self.H)

            return samples_z, jxz

    def _transform_z2x(self, samples_z, jacobian=False):
        """
        This is a method to transform a standard normal vector :math:`\mathbf{z}` to a vector
        :math:`\mathbf{x}` of samples with marginal distributions :math:`f_i(x_i)` and cumulative distributions
        :math:`F_i(x_i)` to samples  according to: :math:`Z_{i}=\Phi^{-1}(F_i(X_{i}))`, where :math:`\Phi` is the
        cumulative distribution function of a standard  normal variable.

        This method is part of the ``Nataf`` class.

        **Inputs:**

        * **samples_z** (`ndarray`):
            Standard normal random vector of shape ``(nsamples, dimension)``

        * **jacobian** (`Boolean`):
            A boolean whether to return the jacobian of the transformation.

            Default: False

        **Outputs:**

        * **samples_x** (`ndarray`):
            Random vector of shape ``(nsamples, dimension)`` with prescribed probability distributions.

        * **jzx** (`ndarray`):
            The jacobian of the transformation of shape ``(dimension, dimension)``.

        """

        m, n = np.shape(samples_z)
        from scipy.linalg import cholesky
        h = cholesky(self.corr_z, lower=True)
        # samples_z = (h @ samples_y.T).T

        samples_x = np.zeros_like(samples_z)
        if isinstance(self.dist_object, JointInd):
            if all(hasattr(m, 'icdf') for m in self.dist_object.marginals):
                for j in range(len(self.dist_object.marginals)):
                    samples_x[:, j] = self.dist_object.marginals[j].icdf(stats.norm.cdf(samples_z[:, j]))

        elif isinstance(self.dist_object, DistributionContinuous1D):
            samples_x = self.dist_object.icdf(stats.norm.cdf(samples_z))
        elif isinstance(self.dist_object, list):
            for j in range(samples_x.shape[1]):
                samples_x[:, j] = self.dist_object[j].icdf(stats.norm.cdf(samples_z[:, j]))

        if not jacobian:
            return samples_x
        else:
            jac = np.zeros(shape=(n, n))
            jzx = [None] * m
            for i in range(m):
                for j in range(n):
                    xi = np.array([samples_x[i, j]])
                    zi = np.array([samples_z[i, j]])
                    jac[j, j] = self.dist_object[j].pdf(xi) / stats.norm.pdf(zi)
                jzx[i] = np.linalg.solve(h, jac)

            return samples_x, jzx

    def rvs(self, nsamples):
        """
        Generate realizations from the joint pdf of the random vector **X**.

        **Inputs:**

        * **nsamples** (`int`):
            Number of samples to generate.

        **Outputs:**

        * **samples_x** (`ndarray`):
            Random vector in the parameter space of shape ``(nsamples, dimension)``.

        """
        from scipy.linalg import cholesky
        h = cholesky(self.corr_z, lower=True)
        n = int(nsamples)
        m = np.size(self.dist_object)
        y = np.random.randn(nsamples, m)
        z = np.dot(h, y.T).T
        samples_x = np.zeros([n, m])
        for i in range(m):
            samples_x[:, i] = self.dist_object[i].icdf(stats.norm.cdf(z[:, i]))
        return samples_x





import numpy as np

from UQpy.Inference.InferenceModel import InferenceModel
from UQpy.Inference.MLEstimation import MLEstimation

########################################################################################################################
########################################################################################################################
#                                  Model Selection Using Information Theoretic Criteria
########################################################################################################################

class InfoModelSelection:
    """
    Perform model selection using information theoretic criteria.

    Supported criteria are BIC, AIC (default), AICc. This class leverages the ``MLEstimation`` class for maximum
    likelihood estimation, thus inputs to ``MLEstimation`` can also be provided to ``InfoModelSelection``, as lists of
    length equal to the number of models.


    **Inputs:**

    * **candidate_models** (`list` of ``InferenceModel`` objects):
        Candidate models

    * **data** (`ndarray`):
        Available data

    * **criterion** (`str`):
        Criterion to be used ('AIC', 'BIC', 'AICc'). Default is 'AIC'

    * **kwargs**:
        Additional keyword inputs to the maximum likelihood estimators.

        Keys must refer to input names to the ``MLEstimation`` class, and values must be lists of length `nmodels`,
        ordered in the same way as input `candidate_models`. For example, setting
        `kwargs={`method': [`Nelder-Mead', `Powell']}` means that the Nelder-Mead minimization algorithm will be used
        for ML estimation of the first candidate model, while the Powell method will be used for the second candidate
        model.

    * **random_state** (None or `int` or ``numpy.random.RandomState`` object):
        Random seed used to initialize the pseudo-random number generator. Default is None.

        If an integer is provided, this sets the seed for an object of ``numpy.random.RandomState``. Otherwise, the
        object itself can be passed directly.

    * **x0** (`list` of `ndarrays`):
        Starting points for optimization - see ``MLEstimation``

    * **nopt** (`list` of `int`):
        Number of iterations for the maximization procedure - see ``MLEstimation``

    If `x0` and `nopt` are both `None`, the object is created but the model selection procedure is not run, one
    must then call the ``run`` method.

    **Attributes:**

    * **ml_estimators** (`list` of `MLEstimation` objects):
        ``MLEstimation`` results for each model (contains e.g. fitted parameters)

    * **criterion_values** (`list` of `floats`):
        Value of the criterion for all models.

    * **penalty_terms** (`list` of `floats`):
        Value of the penalty term for all models. Data fit term is then criterion_value - penalty_term.

    * **probabilities** (`list` of `floats`):
        Value of the model probabilities, computed as

        .. math:: P(M_i|d) = \dfrac{\exp(-\Delta_i/2)}{\sum_i \exp(-\Delta_i/2)}

        where :math:`\Delta_i = criterion_i - min_i(criterion)`

    **Methods:**

    """
    # Authors: Audrey Olivier, Dimitris Giovanis
    # Last Modified: 12/19 by Audrey Olivier
    def __init__(self, candidate_models, data, criterion='AIC', random_state=None, verbose=False, nopt=None, x0=None,
                 **kwargs):

        # Check inputs
        # candidate_models is a list of InferenceModel objects
        if not isinstance(candidate_models, (list, tuple)) or not all(isinstance(model, InferenceModel)
                                                                      for model in candidate_models):
            raise TypeError('UQpy: Input candidate_models must be a list of InferenceModel objects.')
        self.nmodels = len(candidate_models)
        self.candidate_models = candidate_models
        self.data = data
        if criterion not in ['AIC', 'BIC', 'AICc']:
            raise ValueError('UQpy: Criterion should be AIC (default), BIC or AICc')
        self.criterion = criterion
        self.random_state = random_state
        if isinstance(self.random_state, int):
            self.random_state = np.random.RandomState(self.random_state)
        elif not isinstance(self.random_state, (type(None), np.random.RandomState)):
            raise TypeError('UQpy: random_state must be None, an int or an np.random.RandomState object.')
        self.verbose = verbose

        # Instantiate the ML estimators
        if not all(isinstance(value, (list, tuple)) for (key, value) in kwargs.items()) or \
                not all(len(value) == len(candidate_models) for (key, value) in kwargs.items()):
            raise TypeError('UQpy: Extra inputs to model selection must be lists of length len(candidate_models)')
        self.ml_estimators = []
        for i, inference_model in enumerate(self.candidate_models):
            kwargs_i = dict([(key, value[i]) for (key, value) in kwargs.items()])
            ml_estimator = MLEstimation(inference_model=inference_model, data=self.data, verbose=self.verbose,
                                        random_state=self.random_state, x0=None, nopt=None, **kwargs_i)
            self.ml_estimators.append(ml_estimator)

        # Initialize the outputs
        self.criterion_values = [None, ] * self.nmodels
        self.penalty_terms = [None, ] * self.nmodels
        self.probabilities = [None, ] * self.nmodels

        # Run the model selection procedure
        if (nopt is not None) or (x0 is not None):
            self.run(nopt=nopt, x0=x0)

    def run(self, nopt=1, x0=None):
        """
        Run the model selection procedure, i.e. compute criterion value for all models.

        This function calls the ``run`` method of the ``MLEstimation`` object for each model to compute the maximum
        log-likelihood, then computes the criterion value and probability for each model.

        **Inputs:**

        * **x0** (`list` of `ndarrays`):
            Starting point(s) for optimization for all models. Default is `None`. If not provided, see `nopt`. See
            ``MLEstimation`` class.

        * **nopt** (`int` or `list` of `ints`):
            Number of iterations that the optimization is run, starting at random initial guesses. It is only used if
            `x0` is not provided. Default is 1. See ``MLEstimation`` class.

        """
        # Check inputs x0, nopt
        if isinstance(nopt, int) or nopt is None:
            nopt = [nopt] * self.nmodels
        if not (isinstance(nopt, list) and len(nopt) == self.nmodels):
            raise ValueError('UQpy: nopt should be an int or list of length nmodels')
        if x0 is None:
            x0 = [None] * self.nmodels
        if not (isinstance(x0, list) and len(x0) == self.nmodels):
            raise ValueError('UQpy: x0 should be a list of length nmodels (or None).')

        # Loop over all the models
        for i, (inference_model, ml_estimator) in enumerate(zip(self.candidate_models, self.ml_estimators)):
            # First evaluate ML estimate for all models, do several iterations if demanded
            ml_estimator.run(nopt=nopt[i], x0=x0[i])

            # Then minimize the criterion
            self.criterion_values[i], self.penalty_terms[i] = self._compute_info_criterion(
                criterion=self.criterion, data=self.data, inference_model=inference_model,
                max_log_like=ml_estimator.max_log_like, return_penalty=True)

        # Compute probabilities from criterion values
        self.probabilities = self._compute_probabilities(self.criterion_values)

    def sort_models(self):
        """
        Sort models in descending order of model probability (increasing order of `criterion` value).

        This function sorts - in place - the attribute lists `candidate_models, ml_estimators, criterion_values,
        penalty_terms` and `probabilities` so that they are sorted from most probable to least probable model. It is a
        stand-alone function that is provided to help the user to easily visualize which model is the best.

        No inputs/outputs.

        """
        sort_idx = list(np.argsort(np.array(self.criterion_values)))

        self.candidate_models = [self.candidate_models[i] for i in sort_idx]
        self.ml_estimators = [self.ml_estimators[i] for i in sort_idx]
        self.criterion_values = [self.criterion_values[i] for i in sort_idx]
        self.penalty_terms = [self.penalty_terms[i] for i in sort_idx]
        self.probabilities = [self.probabilities[i] for i in sort_idx]

    @staticmethod
    def _compute_info_criterion(criterion, data, inference_model, max_log_like, return_penalty=False):
        """
        Compute the criterion value for a given model, given a max_log_likelihood value.

        The criterion value is -2 * max_log_like + penalty, the penalty depends on the chosen criterion. This function
        is a utility function (static method), called within the run_estimation method.

        **Inputs:**

        :param criterion: Chosen criterion.
        :type criterion: str

        :param data: Available data.
        :type data: ndarray

        :param inference_model: Inference model.
        :type inference_model: object of class InferenceModel

        :param max_log_like: Value of likelihood function at MLE.
        :type max_log_like: float

        :param return_penalty: Boolean that sets whether to return the penalty term as additional output.

                               Default is False
        :type return_penalty: bool

        **Output/Returns:**

        :return criterion_value: Value of criterion.
        :rtype criterion_value: float

        :return penalty_term: Value of penalty term.
        :rtype penalty_term: float

        """

        n_params = inference_model.nparams
        ndata = len(data)
        if criterion == 'BIC':
            penalty_term = np.log(ndata) * n_params
        elif criterion == 'AICc':
            penalty_term = 2 * n_params + (2 * n_params ** 2 + 2 * n_params) / (ndata - n_params - 1)
        elif criterion == 'AIC':  # default
            penalty_term = 2 * n_params
        else:
            raise ValueError('UQpy: Criterion should be AIC (default), BIC or AICc')
        if return_penalty:
            return -2 * max_log_like + penalty_term, penalty_term
        return -2 * max_log_like + penalty_term

    @staticmethod
    def _compute_probabilities(criterion_values):
        """
        Compute the model probability given criterion values for all models.

        Model probability is proportional to exp(-criterion/2), model probabilities over all models sum up to 1. This
        function is a utility function (static method), called within the run_estimation method.

        **Inputs:**

        :param criterion_values: Values of criterion for all models.
        :type criterion_values: list (length nmodels) of floats

        **Output/Returns:**

        :return probabilities: Values of model probabilities
        :rtype probabilities: list (length nmodels) of floats

        """

        delta = np.array(criterion_values) - min(criterion_values)
        prob = np.exp(-delta / 2)
        return prob / np.sum(prob)








"""Distributions module."""
from UQpy.Distributions.collection.beta import Beta
from UQpy.Distributions.collection.binomial import Binomial
from UQpy.Distributions.collection.cauchy import Cauchy
from UQpy.Distributions.collection.chi_square import ChiSquare
from UQpy.Distributions.collection.exponential import Exponential
from UQpy.Distributions.collection.gamma import Gamma
from UQpy.Distributions.collection.gen_extreme import GenExtreme
from UQpy.Distributions.collection.inverse_gauss import InvGauss
from UQpy.Distributions.collection.laplace import Laplace
from UQpy.Distributions.collection.levy import Levy
from UQpy.Distributions.collection.logistic import Logistic
from UQpy.Distributions.collection.lognormal import Lognormal
from UQpy.Distributions.collection.maxwell import Maxwell
from UQpy.Distributions.collection.multinomial import Multinomial
from UQpy.Distributions.collection.mvnormal import MVNormal
from UQpy.Distributions.collection.normal import Normal
from UQpy.Distributions.collection.pareto import Pareto
from UQpy.Distributions.collection.poisson import Poisson
from UQpy.Distributions.collection.rayleigh import Rayleigh
from UQpy.Distributions.collection.truncated_normal import TruncNorm
from UQpy.Distributions.collection.uniform import Uniform
from UQpy.Distributions.collection.joint_ind import JointInd
from UQpy.Distributions.collection.joint_copula import JointCopula



from UQpy.Distributions.baseclass import Copula
from UQpy.Distributions.baseclass import DistributionContinuous1D, DistributionND, DistributionDiscrete1D
import numpy as np


class Clayton(Copula):
    """
    Clayton copula having cumulative distribution function

    .. math:: F(u_1, u_2) = \max(u_1^{-\Theta} + u_2^{-\Theta} - 1, 0)^{-1/{\Theta}}

    where :math:`u_1 = F_1(x_1), u_2 = F_2(x_2)` are uniformly distributed on the interval `[0, 1]`.

    **Input:**

    * **theta** (`float`):
        Parameter of the copula, real number in [-1, +oo)\{0}.

    This copula possesses the following methods:

    * ``evaluate_cdf`` and ``check_copula``

    (``check_copula`` checks that `marginals` consist of solely 2 continuous univariate distributions).
    """
    def __init__(self, theta):
        # Check the input copula_params
        if theta is not None and ((not isinstance(theta, (float, int))) or (theta < -1 or theta == 0.)):
            raise ValueError('Input theta should be a float in [-1, +oo)\{0}.')
        super().__init__(theta=theta)

    def evaluate_cdf(self, unif):
        if unif.shape[1] > 2:
            raise ValueError('Maximum dimension for the Clayton Copula is 2.')

        #if self.params['theta'] == 1:
        #    return np.prod(unif, axis=1)

        u = unif[:, 0]
        v = unif[:, 1]
        theta = self.params['theta']
        cdf_val = (np.maximum(u ** (-theta) + v ** (-theta) - 1., 0.)) ** (-1. / theta)
        return cdf_val

    @staticmethod
    def check_marginals(marginals):
        if len(marginals) != 2:
            raise ValueError('Maximum dimension for the Clayton Copula is 2.')
        if not all(isinstance(m, DistributionContinuous1D) for m in marginals):
            raise ValueError('Marginals should be 1d continuous distributions.')



import numpy as np
from UQpy.Distributions import Distribution

class MCMC:
    """
    Generate samples from arbitrary user-specified probability density function using Markov Chain Monte Carlo.

    This is the parent class for all MCMC algorithms. This parent class only provides the framework for MCMC and cannot
    be used directly for sampling. Sampling is done by calling the child class for the specific MCMC algorithm.


    **Inputs:**

    * **dimension** (`int`):
        A scalar value defining the dimension of target density function. Either `dimension` and `nchains` or `seed`
        must be provided.

    * **pdf_target** ((`list` of) callables):
        Target density function from which to draw random samples. Either `pdf_target` or `log_pdf_target` must be
        provided (the latter should be preferred for better numerical stability).

        If `pdf_target` is a callable, it refers to the joint pdf to sample from, it must take at least one input `x`,
        which are the point(s) at which to evaluate the pdf. Within MCMC the `pdf_target` is evaluated as:
        ``p(x) = pdf_target(x, *args_target)``

        where `x` is a ndarray of shape (nsamples, dimension) and `args_target` are additional positional arguments that
        are provided to MCMC via its `args_target` input.

        If `pdf_target` is a list of callables, it refers to independent marginals to sample from. The marginal in
        dimension `j` is evaluated as: ``p_j(xj) = pdf_target[j](xj, *args_target[j])`` where `x` is a ndarray of shape
        (nsamples, dimension)

    * **log_pdf_target** ((`list` of) callables):
        Logarithm of the target density function from which to draw random samples. Either `pdf_target` or
        `log_pdf_target` must be provided (the latter should be preferred for better numerical stability).

        Same comments as for input `pdf_target`.

    * **args_target** ((`list` of) `tuple`):
        Positional arguments of the pdf / log-pdf target function. See `pdf_target`

    * **seed** (`ndarray`):
        Seed of the Markov chain(s), shape ``(nchains, dimension)``. Default: zeros(`nchains` x `dimension`).

        If `seed` is not provided, both `nchains` and `dimension` must be provided.

    * **nburn** (`int`):
        Length of burn-in - i.e., number of samples at the beginning of the chain to discard (note: no thinning during
        burn-in). Default is 0, no burn-in.

    * **jump** (`int`):
        Thinning parameter, used to reduce correlation between samples. Setting `jump=n` corresponds to	skipping `n-1`
        states between accepted states of the chain. Default is 1 (no thinning).

    * **nchains** (`int`):
        The number of Markov chains to generate. Either `dimension` and `nchains` or `seed` must be provided.

    * **save_log_pdf** (`bool`):
        Boolean that indicates whether to save log-pdf values along with the samples. Default: False

    * **verbose** (`boolean`)
        Set ``verbose = True`` to print status messages to the terminal during execution.

    * **concat_chains** (`bool`):
        Boolean that indicates whether to concatenate the chains after a run, i.e., samples are stored as an `ndarray`
        of shape (nsamples * nchains, dimension) if True, (nsamples, nchains, dimension) if False. Default: True

    * **random_state** (None or `int` or ``numpy.random.RandomState`` object):
        Random seed used to initialize the pseudo-random number generator. Default is None.

        If an integer is provided, this sets the seed for an object of ``numpy.random.RandomState``. Otherwise, the
        object itself can be passed directly.


    **Attributes:**

    * **samples** (`ndarray`)
        Set of MCMC samples following the target distribution, `ndarray` of shape (`nsamples` * `nchains`, `dimension`)
        or (nsamples, nchains, dimension) (see input `concat_chains`).

    * **log_pdf_values** (`ndarray`)
        Values of the log pdf for the accepted samples, `ndarray` of shape (nchains * nsamples,) or (nsamples, nchains)

    * **nsamples** (`list`)
        Total number of samples; The `nsamples` attribute tallies the total number of generated samples. After each
        iteration, it is updated by 1. At the end of the simulation, the `nsamples` attribute equals the user-specified
        value for input `nsamples` given to the child class.

    * **nsamples_per_chain** (`list`)
        Total number of samples per chain; Similar to the attribute `nsamples`, it is updated during iterations as new
        samples are saved.

    * **niterations** (`list`)
        Total number of iterations, updated on-the-fly as the algorithm proceeds. It is related to number of samples as
        niterations=nburn+jump*nsamples_per_chain.

    * **acceptance_rate** (`list`)
        Acceptance ratio of the MCMC chains, computed separately for each chain.

    **Methods:**
    """
    # Last Modified: 10/05/20 by Audrey Olivier

    def __init__(self, dimension=None, pdf_target=None, log_pdf_target=None, args_target=None, seed=None, nburn=0,
                 jump=1, nchains=None, save_log_pdf=False, verbose=False, concat_chains=True, random_state=None):

        if not (isinstance(nburn, int) and nburn >= 0):
            raise TypeError('UQpy: nburn should be an integer >= 0')
        if not (isinstance(jump, int) and jump >= 1):
            raise TypeError('UQpy: jump should be an integer >= 1')
        self.nburn, self.jump = nburn, jump
        self.seed = self._preprocess_seed(seed=seed, dim=dimension, nchains=nchains)
        self.nchains, self.dimension = self.seed.shape

        # Check target pdf
        self.evaluate_log_target, self.evaluate_log_target_marginals = self._preprocess_target(
            pdf_=pdf_target, log_pdf_=log_pdf_target, args=args_target)
        self.save_log_pdf = save_log_pdf
        self.concat_chains = concat_chains
        self.random_state = random_state
        if isinstance(self.random_state, int):
            self.random_state = np.random.RandomState(self.random_state)
        elif not isinstance(self.random_state, (type(None), np.random.RandomState)):
            raise TypeError('UQpy: random_state must be None, an int or an np.random.RandomState object.')
        self.verbose = verbose

        self.log_pdf_target = log_pdf_target
        self.pdf_target = pdf_target
        self.args_target = args_target

        # Initialize a few more variables
        self.samples = None
        self.log_pdf_values = None
        self.acceptance_rate = [0.] * self.nchains
        self.nsamples, self.nsamples_per_chain = 0, 0
        self.niterations = 0  # total nb of iterations, grows if you call run several times

    def run(self, nsamples=None, nsamples_per_chain=None):
        """
        Run the MCMC algorithm.

        This function samples from the MCMC chains and appends samples to existing ones (if any). This method leverages
        the ``run_iterations`` method that is specific to each algorithm.

        **Inputs:**

        * **nsamples** (`int`):
            Number of samples to generate.

        * **nsamples_per_chain** (`int`)
            Number of samples to generate per chain.

        Either `nsamples` or `nsamples_per_chain` must be provided (not both). Not that if `nsamples` is not a multiple
        of `nchains`, `nsamples` is set to the next largest integer that is a multiple of `nchains`.

        """
        # Initialize the runs: allocate space for the new samples and log pdf values
        final_nsamples, final_nsamples_per_chain, current_state, current_log_pdf = self._initialize_samples(
            nsamples=nsamples, nsamples_per_chain=nsamples_per_chain)

        if self.verbose:
            print('UQpy: Running MCMC...')

        # Run nsims iterations of the MCMC algorithm, starting at current_state
        while self.nsamples_per_chain < final_nsamples_per_chain:
            # update the total number of iterations
            self.niterations += 1
            # run iteration
            current_state, current_log_pdf = self.run_one_iteration(current_state, current_log_pdf)
            # Update the chain, only if burn-in is over and the sample is not being jumped over
            # also increase the current number of samples and samples_per_chain
            if self.niterations > self.nburn and (self.niterations - self.nburn) % self.jump == 0:
                self.samples[self.nsamples_per_chain, :, :] = current_state.copy()
                if self.save_log_pdf:
                    self.log_pdf_values[self.nsamples_per_chain, :] = current_log_pdf.copy()
                self.nsamples_per_chain += 1
                self.nsamples += self.nchains

        if self.verbose:
            print('UQpy: MCMC run successfully !')

        # Concatenate chains maybe
        if self.concat_chains:
            self._concatenate_chains()

    def run_one_iteration(self, current_state, current_log_pdf):
        """
        Run one iteration of the MCMC algorithm, starting at `current_state`.

        This method is over-written for each different MCMC algorithm. It must return the new state and associated
        log-pdf, which will be passed as inputs to the ``run_one_iteration`` method at the next iteration.

        **Inputs:**

        * **current_state** (`ndarray`):
            Current state of the chain(s), `ndarray` of shape ``(nchains, dimension)``.

        * **current_log_pdf** (`ndarray`):
            Log-pdf of the current state of the chain(s), `ndarray` of shape ``(nchains, )``.

        **Outputs/Returns:**

        * **new_state** (`ndarray`):
            New state of the chain(s), `ndarray` of shape ``(nchains, dimension)``.

        * **new_log_pdf** (`ndarray`):
            Log-pdf of the new state of the chain(s), `ndarray` of shape ``(nchains, )``.

        """
        return [], []

    ####################################################################################################################
    # Helper functions that can be used by all algorithms
    # Methods update_samples, update_accept_ratio and sample_candidate_from_proposal can be called in the run stage.
    # Methods preprocess_target, preprocess_proposal, check_seed and check_integers can be called in the init stage.

    def _concatenate_chains(self):
        """
        Concatenate chains.

        Utility function that reshapes (in place) attribute samples from (nsamples, nchains, dimension) to
        (nsamples * nchains, dimension), and log_pdf_values from (nsamples, nchains) to (nsamples * nchains, ).

        No input / output.

        """
        self.samples = self.samples.reshape((-1, self.dimension), order='C')
        if self.save_log_pdf:
            self.log_pdf_values = self.log_pdf_values.reshape((-1, ), order='C')
        return None

    def _unconcatenate_chains(self):
        """
        Inverse of concatenate_chains.

        Utility function that reshapes (in place) attribute samples from (nsamples * nchains, dimension) to
        (nsamples, nchains, dimension), and log_pdf_values from (nsamples * nchains) to (nsamples, nchains).

        No input / output.

        """
        self.samples = self.samples.reshape((-1, self.nchains, self.dimension), order='C')
        if self.save_log_pdf:
            self.log_pdf_values = self.log_pdf_values.reshape((-1, self.nchains), order='C')
        return None

    def _initialize_samples(self, nsamples, nsamples_per_chain):
        """
        Initialize necessary attributes and variables before running the chain forward.

        Utility function that allocates space for samples and log likelihood values, initialize sample_index,
        acceptance ratio. If some samples already exist, allocate space to append new samples to the old ones. Computes
        the number of forward iterations nsims to be run (depending on burnin and jump parameters).

        **Inputs:**

        * nchains (int): number of chains run in parallel
        * nsamples (int): number of samples to be generated
        * nsamples_per_chain (int): number of samples to be generated per chain

        **Output/Returns:**

        * nsims (int): Number of iterations to perform
        * current_state (ndarray of shape (nchains, dim)): Current state of the chain to start from.

        """
        if ((nsamples is not None) and (nsamples_per_chain is not None)) or (
                nsamples is None and nsamples_per_chain is None):
            raise ValueError('UQpy: Either nsamples or nsamples_per_chain must be provided (not both)')
        if nsamples_per_chain is not None:
            if not (isinstance(nsamples_per_chain, int) and nsamples_per_chain >= 0):
                raise TypeError('UQpy: nsamples_per_chain must be an integer >= 0.')
            nsamples = int(nsamples_per_chain * self.nchains)
        else:
            if not (isinstance(nsamples, int) and nsamples >= 0):
                raise TypeError('UQpy: nsamples must be an integer >= 0.')
            nsamples_per_chain = int(np.ceil(nsamples / self.nchains))
            nsamples = int(nsamples_per_chain * self.nchains)

        if self.samples is None:    # very first call of run, set current_state as the seed and initialize self.samples
            self.samples = np.zeros((nsamples_per_chain, self.nchains, self.dimension))
            if self.save_log_pdf:
                self.log_pdf_values = np.zeros((nsamples_per_chain, self.nchains))
            current_state = np.zeros_like(self.seed)
            np.copyto(current_state, self.seed)
            current_log_pdf = self.evaluate_log_target(current_state)
            if self.nburn == 0:    # if nburn is 0, save the seed, run one iteration less 
                self.samples[0, :, :] = current_state
                if self.save_log_pdf:
                    self.log_pdf_values[0, :] = current_log_pdf
                self.nsamples_per_chain += 1
                self.nsamples += self.nchains
            final_nsamples, final_nsamples_per_chain = nsamples, nsamples_per_chain

        else:    # fetch previous samples to start the new run, current state is last saved sample
            if len(self.samples.shape) == 2:   # the chains were previously concatenated
                self._unconcatenate_chains()
            current_state = self.samples[-1]
            current_log_pdf = self.evaluate_log_target(current_state)
            self.samples = np.concatenate(
                [self.samples, np.zeros((nsamples_per_chain, self.nchains, self.dimension))], axis=0)
            if self.save_log_pdf:
                self.log_pdf_values = np.concatenate(
                    [self.log_pdf_values, np.zeros((nsamples_per_chain, self.nchains))], axis=0)
            final_nsamples = nsamples + self.nsamples
            final_nsamples_per_chain = nsamples_per_chain + self.nsamples_per_chain

        return final_nsamples, final_nsamples_per_chain, current_state, current_log_pdf

    def _update_acceptance_rate(self, new_accept=None):
        """
        Update acceptance rate of the chains.

        Utility function, uses an iterative function to update the acceptance rate of all the chains separately.

        **Inputs:**

        * new_accept (list (length nchains) of bool): indicates whether the current state was accepted (for each chain
          separately).

        """
        self.acceptance_rate = [na / self.niterations + (self.niterations - 1) / self.niterations * a
                                for (na, a) in zip(new_accept, self.acceptance_rate)]

    @staticmethod
    def _preprocess_target(log_pdf_, pdf_, args):
        """
        Preprocess the target pdf inputs.

        Utility function (static method), that transforms the log_pdf, pdf, args inputs into a function that evaluates
        log_pdf_target(x) for a given x. If the target is given as a list of callables (marginal pdfs), the list of
        log margianals is also returned.

        **Inputs:**

        * log_pdf_ ((list of) callables): Log of the target density function from which to draw random samples. Either
          pdf_target or log_pdf_target must be provided.
        * pdf_ ((list of) callables): Target density function from which to draw random samples. Either pdf_target or
          log_pdf_target must be provided.
        * args (tuple): Positional arguments of the pdf target.

        **Output/Returns:**

        * evaluate_log_pdf (callable): Callable that computes the log of the target density function
        * evaluate_log_pdf_marginals (list of callables): List of callables to compute the log pdf of the marginals

        """
        # log_pdf is provided
        if log_pdf_ is not None:
            if callable(log_pdf_):
                if args is None:
                    args = ()
                evaluate_log_pdf = (lambda x: log_pdf_(x, *args))
                evaluate_log_pdf_marginals = None
            elif isinstance(log_pdf_, list) and (all(callable(p) for p in log_pdf_)):
                if args is None:
                    args = [()] * len(log_pdf_)
                if not (isinstance(args, list) and len(args) == len(log_pdf_)):
                    raise ValueError('UQpy: When log_pdf_target is a list, args should be a list (of tuples) of same '
                                     'length.')
                evaluate_log_pdf_marginals = list(
                    map(lambda i: lambda x: log_pdf_[i](x, *args[i]), range(len(log_pdf_))))
                evaluate_log_pdf = (lambda x: np.sum(
                    [log_pdf_[i](x[:, i, np.newaxis], *args[i]) for i in range(len(log_pdf_))]))
            else:
                raise TypeError('UQpy: log_pdf_target must be a callable or list of callables')
        # pdf is provided
        elif pdf_ is not None:
            if callable(pdf_):
                if args is None:
                    args = ()
                evaluate_log_pdf = (lambda x: np.log(np.maximum(pdf_(x, *args), 10 ** (-320) * np.ones((x.shape[0],)))))
                evaluate_log_pdf_marginals = None
            elif isinstance(pdf_, (list, tuple)) and (all(callable(p) for p in pdf_)):
                if args is None:
                    args = [()] * len(pdf_)
                if not (isinstance(args, (list, tuple)) and len(args) == len(pdf_)):
                    raise ValueError('UQpy: When pdf_target is given as a list, args should also be a list of same '
                                     'length.')
                evaluate_log_pdf_marginals = list(
                    map(lambda i: lambda x: np.log(np.maximum(pdf_[i](x, *args[i]),
                                                              10 ** (-320) * np.ones((x.shape[0],)))),
                        range(len(pdf_))
                        ))
                evaluate_log_pdf = (lambda x: np.sum(
                    [np.log(np.maximum(pdf_[i](x[:, i, np.newaxis], *args[i]), 10**(-320)*np.ones((x.shape[0],))))
                     for i in range(len(log_pdf_))]))
            else:
                raise TypeError('UQpy: pdf_target must be a callable or list of callables')
        else:
            raise ValueError('UQpy: log_pdf_target or pdf_target should be provided.')
        return evaluate_log_pdf, evaluate_log_pdf_marginals

    @staticmethod
    def _preprocess_seed(seed, dim, nchains):
        """
        Preprocess input seed.

        Utility function (static method), that checks the dimension of seed, assign [0., 0., ..., 0.] if not provided.

        **Inputs:**

        * seed (ndarray): seed for MCMC
        * dim (int): dimension of target density

        **Output/Returns:**

        * seed (ndarray): seed for MCMC
        * dim (int): dimension of target density

        """
        if seed is None:
            if dim is None or nchains is None:
                raise ValueError('UQpy: Either `seed` or `dimension` and `nchains` must be provided.')
            seed = np.zeros((nchains, dim))
        else:
            seed = np.atleast_1d(seed)
            if len(seed.shape) == 1:
                seed = np.reshape(seed, (1, -1))
            elif len(seed.shape) > 2:
                raise ValueError('UQpy: Input seed should be an array of shape (dimension, ) or (nchains, dimension).')
            if dim is not None and seed.shape[1] != dim:
                raise ValueError('UQpy: Wrong dimensions between seed and dimension.')
            if nchains is not None and seed.shape[0] != nchains:
                raise ValueError('UQpy: The number of chains and the seed shape are inconsistent.')
        return seed

    @staticmethod
    def _check_methods_proposal(proposal):
        """
        Check if proposal has required methods.

        Utility function (static method), that checks that the given proposal distribution has 1) a rvs method and 2) a
        log pdf or pdf method. If a pdf method exists but no log_pdf, the log_pdf methods is added to the proposal
        object. Used in the MH and MMH initializations.

        **Inputs:**

        * proposal (Distribution object): proposal distribution

        """
        if not isinstance(proposal, Distribution):
            raise TypeError('UQpy: Proposal should be a Distribution object')
        if not hasattr(proposal, 'rvs'):
            raise AttributeError('UQpy: The proposal should have an rvs method')
        if not hasattr(proposal, 'log_pdf'):
            if not hasattr(proposal, 'pdf'):
                raise AttributeError('UQpy: The proposal should have a log_pdf or pdf method')
            proposal.log_pdf = lambda x: np.log(np.maximum(proposal.pdf(x), 10 ** (-320) * np.ones((x.shape[0],))))





import numpy as np
from UQpy.Distributions import DistributionContinuous1D

########################################################################################################################
########################################################################################################################
#                                         Stochastic Reduced Order Model  (SROM)                                       #
########################################################################################################################
########################################################################################################################


class SROM:

    """
    Stochastic Reduced Order Model(SROM) provide a low-dimensional, discrete approximation of a given random
    quantity.

    **Inputs:**

    * **samples** (`ndarray`):
        An array/list of samples corresponding to the points at which the SROM is defined.

    * **target_dist_object** ((list of) ``Distribution`` object(s)):
        A list of distribution objects for each random variable.

    * **moments** (`list` of `float`):
        A list containing first and second order moment about origin of all random variables.

    * **weights_errors** (`list` of `float`):
        A list of weights associated with the error in distribution, moments and correlation.

        This corresponds to a list of the values :math:`a_{u}` in the objective function above.

        Default: weights_errors = [1, 0.2, 0]

    * **properties** (`list` of `booleans`):
        A list of booleans declaring the properties to be matched in the reduced order model.

        `properties[0] = True` matches the marginal distributions

        `properties[1] = True` matches the mean values

        `properties[2] = True` matches the mean square

        `properties[3] = True` matches the correlation

    * **weights_distribution** (`ndarray` or `list` of `float`):
        A list or array containing weights associated with matching the distribution at each sample value.

        `weights_distribution` is an array or list of shape `(m, d)` where each weight corresponds to the weight
        :math:`w_F(x_{k,i}; i)` assigned for matching the distribution of component `i` at sample point
        :math:`x_{k,i}`.

        If `weights_distribution` is `(1, d)`, it is assumed that each sample sample is equally weighted according
        to the corresponding weight for its distribution.

        Default: `weights_distribution` = An array of shape `(m, d)` with all elements equal to 1.

    * **weights_moments** (`ndarray` or `list` of `float`):
        An list or array containing weights associated with matching the moments about the origin for each
        component.

        `weights_moments` is a list or array of shape `(2, d), where each weight corresponds to the weight
        :math:`w_{\mu}(r; i)` assigned for matching the moment of order :math:`r = 1, 2` for component `i`.

        If `weights_moments` is `(1, d)`, it is assumed that moments of all order are equally weighted.

        Default: `weights_moments` = [[1/(moment[0][i]^2)], [1/(moment[1][i]^2)]] for i = 1, 2, ..., d.

    * **weights_correlation** (`ndarray` or `list` of `float`):
        A list or array containing weights associated with matching the correlation of the random variables.

        `weights_correlation` is a list or array of shape `(d, d)` where each weight corresponds to the weight
        :math:`w_R(i, j)` assigned for matching the correlation between component `i` and component `j`

        Default: `weights_correlation` = `(d, d)` array with all elements equal to 1.

    * **correlation** (`ndarray` or `list of floats`):
        Correlation matrix between random variables.

    **Attribute:**

    * **sample_weights** (`ndarray`):
        The probability weights defining discrete approximation of continuous random variables.

    **Methods:**

    """

    def __init__(self, samples, target_dist_object, moments=None, weights_errors=None, weights_distribution=None,
                 weights_moments=None, weights_correlation=None, properties=None, correlation=None, verbose=False):

        self.target_dist_object = target_dist_object
        self.correlation = correlation
        self.moments = moments

        self.weights_distribution = weights_distribution
        self.weights_moments = weights_moments
        self.weights_correlation = weights_correlation
        self.weights_errors = weights_errors

        self.properties = properties
        self.verbose = verbose
        self.sample_weights = None

        if isinstance(samples, list):
            self.samples = np.array(samples)
            self.nsamples = self.samples.shape[0]
            self.dimension = self.samples.shape[1]
        elif isinstance(samples, np.ndarray):
            self.dimension = samples.shape[1]
            self.samples = samples
            self.nsamples = samples.shape[0]
        else:
            raise NotImplementedError("UQpy: 'samples' sholud be a list or numpy array")

        if self.target_dist_object is None:
            raise NotImplementedError("UQpy: Target Distribution is not defined.")

        if isinstance(self.target_dist_object, list):
            for i in range(len(self.target_dist_object)):
                if not isinstance(self.target_dist_object[i], DistributionContinuous1D):
                    raise TypeError('UQpy: A DistributionContinuous1D object must be provided.')

        if self.properties is not None:
            self.run()
        else:
            print('UQpy: No properties list provided, execute the SROM by calling run method and specifying a '
                  'properties list')

    def run(self, weights_errors=None, weights_distribution=None, weights_moments=None, weights_correlation=None,
            properties=None):
        """
        Execute the stochastic reduced order model in the ``SROM`` class.

        The ``run`` method is the function that computes the probability weights corresponding to the sample. If
        `properties` is provided, the ``run`` method is automatically called when the ``SROM`` object is defined. The
        user may also call the ``run`` method directly to generate samples. The ``run`` method of the ``SROM`` class can
        be invoked many times with different weights parameters and each time computed probability weights are
        overwritten.

        **Inputs:**

        * **weights_errors** (`list` of `float`):
            A list of weights associated with the error in distribution, moments and correlation.

            This corresponds to a list of the values :math:`a_{u}` in the objective function above.

            Default: weights_errors = [1, 0.2, 0]

        * **properties** (`list` of `booleans`):
            A list of booleans declaring the properties to be matched in the reduced order model.

            `properties[0] = True` matches the marginal distributions

            `properties[1] = True` matches the mean values

            `properties[2] = True` matches the mean square

            `properties[3] = True` matches the correlation

        * **weights_distribution** (`ndarray` or `list` of `float`):
            A list or array containing weights associated with matching the distribution at each sample value.

            `weights_distribution` is an array or list of shape `(m, d)` where each weight corresponds to the weight
            :math:`w_F(x_{k,i}; i)` assigned for matching the distribution of component `i` at sample point
            :math:`x_{k,i}`.

            If `weights_distribution` is `(1, d)`, it is assumed that each sample sample is equally weighted according
            to the corresponding weight for its distribution.

            Default: `weights_distribution` = An array of shape `(m, d)` with all elements equal to 1.

        * **weights_moments** (`ndarray` or `list` of `float`):
            An list or array containing weights associated with matching the moments about the origin for each
            component.

            `weights_moments` is a list or array of shape `(2, d), where each weight corresponds to the weight
            :math:`w_{\mu}(r; i)` assigned for matching the moment of order :math:`r = 1, 2` for component `i`.

            If `weights_moments` is `(1, d)`, it is assumed that moments of all order are equally weighted.

            Default: `weights_moments` = [[1/(moment[0][i]^2)], [1/(moment[1][i]^2)]] for i = 1, 2, ..., d.

        * **weights_correlation** (`ndarray` or `list` of `float`):
            A list or array containing weights associated with matching the correlation of the random variables.

            `weights_correlation` is a list or array of shape `(d, d)` where each weight corresponds to the weight
            :math:`w_R(i, j)` assigned for matching the correlation between component `i` and component `j`

            Default: `weights_correlation` = `(d, d)` array with all elements equal to 1.

        """
        from scipy import optimize
        self.weights_distribution = weights_distribution
        self.weights_moments = weights_moments
        self.weights_correlation = weights_correlation
        self.weights_errors = weights_errors
        self.properties = properties

        # Check properties to match
        if self.properties is None:
            self.properties = [True, True, True, False]

        self._init_srom()

        if self.verbose:
            print('UQpy: Performing SROM...')

        def f(p0, samples, wd, wm, wc, mar, n, d, m, alpha, prop, correlation):
            e1 = 0.
            e2 = 0.
            e22 = 0.
            e3 = 0.
            com = np.append(samples, np.atleast_2d(p0).T, 1)
            for j in range(d):
                srt = com[np.argsort(com[:, j].flatten())]
                s = srt[:, j]
                a = srt[:, d]
                a0 = np.cumsum(a)
                marginal = mar[j].cdf

                if prop[0] is True:
                    for i in range(n):
                        e1 += wd[i, j] * (a0[i] - marginal(s[i])) ** 2

                if prop[1] is True:
                    e2 += wm[0, j] * (np.sum(p0 * samples[:, j]) - m[0, j]) ** 2

                if prop[2] is True:
                    e22 += wm[1, j] * (
                            np.sum(np.array(p0) * (samples[:, j] * samples[:, j])) - m[1, j]) ** 2

                if prop[3] is True:
                    for k in range(d):
                        if k > j:
                            r = correlation[j, k] * np.sqrt((m[1, j] - m[0, j] ** 2) * (m[1, k] - m[0, k] ** 2)) + \
                                m[0, j] * m[0, k]
                            e3 += wc[k, j] * (np.sum(p0 * (samples[:, j] * samples[:, k])) - r) ** 2

            return alpha[0] * e1 + alpha[1] * (e2 + e22) + alpha[2] * e3

        def constraint(x):
            return np.sum(x) - 1

        cons = {'type': 'eq', 'fun': constraint}

        p_ = optimize.minimize(f, np.zeros(self.nsamples),
                               args=(self.samples, self.weights_distribution, self.weights_moments,
                                     self.weights_correlation, self.target_dist_object, self.nsamples, self.dimension,
                                     self.moments, self.weights_errors, self.properties, self.correlation),
                               constraints=cons, method='SLSQP', bounds=[[0, 1]]*self.nsamples)

        self.sample_weights = p_.x
        if self.verbose:
            print('UQpy: SROM completed!')

    def _init_srom(self):
        """
        Initialization and preliminary error checks.
        """
        if isinstance(self.moments, list):
            self.moments = np.array(self.moments)

        if isinstance(self.correlation, list):
            self.correlation = np.array(self.correlation)

        # Check moments and correlation
        if self.properties[1] is True or self.properties[2] is True or self.properties[3] is True:
            if self.moments is None:
                raise NotImplementedError("UQpy: 'moments' are required")
        # Both moments are required, if correlation property is required to be match
        if self.properties[3] is True:
            if self.moments.shape != (2, self.dimension):
                raise NotImplementedError("UQpy: Shape of 'moments' is not correct")
            if self.correlation is None:
                self.correlation = np.identity(self.dimension)
        # moments.shape[0] should be 1 or 2
        if self.moments.shape != (1, self.dimension) and self.moments.shape != (2, self.dimension):
            raise NotImplementedError("UQpy: Shape of 'moments' is not correct")
        # If both the moments are to be included in objective function, then moments.shape[0] should be 2
        if self.properties[1] is True and self.properties[2] is True:
            if self.moments.shape != (2, self.dimension):
                raise NotImplementedError("UQpy: Shape of 'moments' is not correct")
        # If only second order moment is to be included in objective function and moments.shape[0] is 1. Then
        # self.moments is converted shape = (2, self.dimension) where is second row contain second order moments.
        if self.properties[1] is False and self.properties[2] is True:
            if self.moments.shape == (1, self.dimension):
                temp = np.ones(shape=(1, self.dimension))
                self.moments = np.concatenate((temp, self.moments))

        # Check weights corresponding to errors
        if self.weights_errors is None:
            self.weights_errors = [1, 0.2, 0]
        elif isinstance(self.weights_errors, list):
            self.weights_errors = np.array(self.weights_errors).astype(np.float64)
        elif not isinstance(self.weights_errors, np.ndarray):
            raise NotImplementedError("UQpy: weights_errors attribute should be a list or numpy array")

        # Check weights corresponding to distribution
        if self.weights_distribution is None:
            self.weights_distribution = np.ones(shape=(self.samples.shape[0], self.dimension))
        elif isinstance(self.weights_distribution, list):
            self.weights_distribution = np.array(self.weights_distribution)
        elif not isinstance(self.weights_distribution, np.ndarray):
            raise NotImplementedError("UQpy: weights_distribution attribute should be a list or numpy array")

        if self.weights_distribution.shape == (1, self.dimension):
            self.weights_distribution = self.weights_distribution * np.ones(shape=(self.samples.shape[0],
                                                                                   self.dimension))
        elif self.weights_distribution.shape != (self.samples.shape[0], self.dimension):
            raise NotImplementedError("UQpy: Size of 'weights for distribution' is not correct")

        # Check weights corresponding to moments and it's default list
        if self.weights_moments is None:
            self.weights_moments = np.reciprocal(np.square(self.moments))
        elif isinstance(self.weights_moments, list):
            self.weights_moments = np.array(self.weights_moments)
        elif not isinstance(self.weights_moments, np.ndarray):
            raise NotImplementedError("UQpy: weights_moments attribute should be a list or numpy array")

        if self.weights_moments.shape == (1, self.dimension):
            self.weights_moments = self.weights_moments * np.ones(shape=(2, self.dimension))
        elif self.weights_moments.shape != (2, self.dimension):
            raise NotImplementedError("UQpy: Size of 'weights for moments' is not correct")

        # Check weights corresponding to correlation and it's default list
        if self.weights_correlation is None:
            self.weights_correlation = np.ones(shape=(self.dimension, self.dimension))
        elif isinstance(self.weights_correlation, list):
            self.weights_correlation = np.array(self.weights_correlation)
        elif not isinstance(self.weights_correlation, np.ndarray):
            raise NotImplementedError("UQpy: weights_correlation attribute should be a list or numpy array")

        if self.weights_correlation.shape != (self.dimension, self.dimension):
            raise NotImplementedError("UQpy: Size of 'weights for correlation' is not correct")


# UQpy is distributed under the MIT license.
#
# Copyright (C) 2018  -- Michael D. Shields
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""This module contains functionality for all the surrogate methods supported in UQpy.

The module currently contains the following classes:

- ``SROM``: Class to estimate a discrete approximation for a continuous random variable using Stochastic Reduced Order
            Model.
- ``Kriging``: Class to generate an approximate surrogate model using Kriging.

- ``PCE``: Class to generate an approximate surrogate model using Polynomial Chaos Expansion.
"""

import numpy as np
import scipy.stats as stats
from UQpy.Distributions import Normal, Uniform, DistributionContinuous1D, JointInd
import scipy.integrate as integrate
import scipy.special as special
import itertools, math
import warnings

warnings.filterwarnings("ignore")

########################################################################################################################
########################################################################################################################
#                                         Kriging Interpolation  (Kriging)                                             #
########################################################################################################################
########################################################################################################################

class Kriging:
    """
    Kriging generates an Gaussian process regression-based surrogate model to predict the model output at new sample
    points.

    **Inputs:**

    * **reg_model** (`str` or `function`):
        `reg_model` specifies and evaluates the basis functions and their coefficients, which defines the trend of
        the model.

        Built-in options (string input): 'Constant', 'Linear', 'Quadratic'

        The user may also pass a callable function as defined in `User-Defined Regression Model` above.

    * **corr_model** (`str` or `function`):
        `corr_model` specifies and evaluates the correlation function.

        Built-in options (string input): 'Exponential', 'Gaussian', 'Linear', 'Spherical', 'Cubic', 'Spline'

        The user may also pass a callable function as defined in `User-Defined Correlation` above.

    * **corr_model_params** (`ndarray` or `list of floats`):
        List or array of initial values for the correlation model hyperparameters/scale parameters.

    * **bounds** (`list` of `float`):
        Bounds on the hyperparameters used to solve optimization problem to estimate maximum likelihood estimator.
        This should be a closed bound.

        Default: [0.001, 10**7] for each hyperparameter.

    * **op** (`boolean`):
        Indicator to solve MLE problem or not. If 'True' corr_model_params will be used as initial solution for
        optimization problem. Otherwise, corr_model_params will be directly use as the hyperparamters.

        Default: True.

    * **nopt** (`int`):
            Number of times MLE optimization problem is to be solved with a random starting point.

            Default: 1.

    * **verbose** (`Boolean`):
            A boolean declaring whether to write text to the terminal.

            Default value: False

    **Attributes:**

    * **beta** (`ndarray`):
            Regression coefficients.

    * **err_var** (`ndarray`):
            Variance of the Gaussian random process.

    * **C_inv** (`ndarray`):
            Inverse Cholesky decomposition of the correlation matrix.

    **Methods:**

    """

    def __init__(self, reg_model='Linear', corr_model='Exponential', bounds=None, op=True, nopt=1, normalize=True,
                 verbose=False, corr_model_params=None, optimizer=None, random_state=None, **kwargs_optimizer):

        self.reg_model = reg_model
        self.corr_model = corr_model
        self.corr_model_params = np.array(corr_model_params)
        self.bounds = bounds
        self.optimizer = optimizer
        self.nopt = nopt
        self.op = op
        self.normalize = normalize
        self.verbose = verbose
        self.random_state = random_state
        self.kwargs_optimizer = kwargs_optimizer

        # Variables are used outside the __init__
        self.samples = None
        self.values = None
        self.sample_mean, self.sample_std = None, None
        self.value_mean, self.value_std = None, None
        self.rmodel, self.cmodel = None, None
        self.beta, self.gamma, self.err_var = None, None, None
        self.F_dash, self.C_inv, self.G = None, None, None
        self.F, self.R = None, None

        # Initialize and run preliminary error checks.
        if self.reg_model is None:
            raise NotImplementedError("UQpy: Regression model is not defined.")

        if self.corr_model is None:
            raise NotImplementedError("Uqpy: Correlation model is not defined.")

        if self.corr_model_params is None:
            raise NotImplementedError("UQpy: corr_model_params is not defined.")

        if self.bounds is None:
            self.bounds = [[0.001, 10 ** 7]] * self.corr_model_params.shape[0]

        if self.optimizer is None:
            from scipy.optimize import fmin_l_bfgs_b
            self.optimizer = fmin_l_bfgs_b
            self.kwargs_optimizer = {'bounds': self.bounds}
        elif not callable(self.optimizer):
            raise TypeError('UQpy: Input optimizer should be None (set to scipy.optimize.minimize) or a callable.')

        if type(self.reg_model).__name__ == 'function':
            self.rmodel = 'User defined'
        elif self.reg_model in ['Constant', 'Linear', 'Quadratic']:
            self.rmodel = self.reg_model
            self.reg_model = self._regress()
        else:
            raise NotImplementedError("UQpy: Doesn't recognize the Regression model.")

        if type(self.corr_model).__name__ == 'function':
            self.cmodel = 'User defined'
        elif self.corr_model in ['Exponential', 'Gaussian', 'Linear', 'Spherical', 'Cubic', 'Spline', 'Other']:
            self.cmodel = self.corr_model
            self.corr_model: callable = self._corr()
        else:
            raise NotImplementedError("UQpy: Doesn't recognize the Correlation model.")

        if isinstance(self.random_state, int):
            self.random_state = np.random.RandomState(self.random_state)
        elif not isinstance(self.random_state, (type(None), np.random.RandomState)):
            raise TypeError('UQpy: random_state must be None, an int or an np.random.RandomState object.')

    def fit(self, samples, values, nopt=None, corr_model_params=None):
        """
        Fit the surrogate model using the training samples and the corresponding model values.

        The user can run this method multiple time after initiating the ``Kriging`` class object.

        This method updates the samples and parameters of the ``Kriging`` object. This method uses `corr_model_params`
        from previous run as the starting point for MLE problem unless user provides a new starting point.

        **Inputs:**

        * **samples** (`ndarray`):
            `ndarray` containing the training points.

        * **values** (`ndarray`):
            `ndarray` containing the model evaluations at the training points.

        **Output/Return:**

        The ``fit`` method has no returns, although it creates the `beta`, `err_var` and `C_inv` attributes of the
        ``Kriging`` class.

        """
        from scipy.linalg import cholesky

        if self.verbose:
            print('UQpy: Running Kriging.fit')

        def log_likelihood(p0, cm, s, f, y):
            # Return the log-likelihood function and it's gradient. Gradient is calculate using Central Difference
            m = s.shape[0]
            n = s.shape[1]
            r__, dr_ = cm(x=s, s=s, params=p0, dt=True)
            try:
                cc = cholesky(r__ + 2 ** (-52) * np.eye(m), lower=True)
            except np.linalg.LinAlgError:
                return np.inf, np.zeros(n)

            # Product of diagonal terms is negligible sometimes, even when cc exists.
            if np.prod(np.diagonal(cc)) == 0:
                return np.inf, np.zeros(n)

            cc_inv = np.linalg.inv(cc)
            r_inv = np.matmul(cc_inv.T, cc_inv)
            f__ = cc_inv.dot(f)
            y__ = cc_inv.dot(y)

            q__, g__ = np.linalg.qr(f__)  # Eq: 3.11, DACE

            # Check if F is a full rank matrix
            if np.linalg.matrix_rank(g__) != min(np.size(f__, 0), np.size(f__, 1)):
                raise NotImplementedError("Chosen regression functions are not sufficiently linearly independent")

            # Design parameters
            beta_ = np.linalg.solve(g__, np.matmul(np.transpose(q__), y__))

            # Computing the process variance (Eq: 3.13, DACE)
            sigma_ = np.zeros(y.shape[1])

            ll = 0
            for out_dim in range(y.shape[1]):
                sigma_[out_dim] = (1 / m) * (np.linalg.norm(y__[:, out_dim] - np.matmul(f__, beta_[:, out_dim])) ** 2)
                # Objective function:= log(det(sigma**2 * R)) + constant
                ll = ll + (np.log(np.linalg.det(sigma_[out_dim] * r__)) + m * (np.log(2 * np.pi) + 1)) / 2

            # Gradient of loglikelihood
            # Reference: C. E. Rasmussen & C. K. I. Williams, Gaussian Processes for Machine Learning, the MIT Press,
            # 2006, ISBN 026218253X. (Page 114, Eq.(5.9))
            residual = y - np.matmul(f, beta_)
            gamma = np.matmul(r_inv, residual)
            grad_mle = np.zeros(n)
            for in_dim in range(n):
                r_inv_derivative = np.matmul(r_inv, np.matmul(dr_[:, :, in_dim], r_inv))
                tmp = np.matmul(residual.T, np.matmul(r_inv_derivative, residual))
                for out_dim in range(y.shape[1]):
                    alpha = gamma / sigma_[out_dim]
                    tmp1 = np.matmul(alpha, alpha.T) - r_inv / sigma_[out_dim]
                    cov_der = sigma_[out_dim] * dr_[:, :, in_dim] + tmp * r__ / m
                    grad_mle[in_dim] = grad_mle[in_dim] - 0.5 * np.trace(np.matmul(tmp1, cov_der))

            return ll, grad_mle

        if nopt is not None:
            self.nopt = nopt
        if corr_model_params is not None:
            self.corr_model_params = corr_model_params
        self.samples = np.array(samples)

        # Number of samples and dimensions of samples and values
        nsamples, input_dim = self.samples.shape
        output_dim = int(np.size(values) / nsamples)

        self.values = np.array(values).reshape(nsamples, output_dim)

        # Normalizing the data
        if self.normalize:
            self.sample_mean, self.sample_std = np.mean(self.samples, 0), np.std(self.samples, 0)
            self.value_mean, self.value_std = np.mean(self.values, 0), np.std(self.values, 0)
            s_ = (self.samples - self.sample_mean) / self.sample_std
            y_ = (self.values - self.value_mean) / self.value_std
        else:
            s_ = self.samples
            y_ = self.values

        self.F, jf_ = self.reg_model(s_)

        # Maximum Likelihood Estimation : Solving optimization problem to calculate hyperparameters
        if self.op:
            starting_point = self.corr_model_params
            minimizer, fun_value = np.zeros([self.nopt, input_dim]), np.zeros([self.nopt, 1])
            for i__ in range(self.nopt):
                p_ = self.optimizer(log_likelihood, starting_point, args=(self.corr_model, s_, self.F, y_),
                                    **self.kwargs_optimizer)
                minimizer[i__, :] = p_[0]
                fun_value[i__, 0] = p_[1]
                # Generating new starting points using log-uniform distribution
                if i__ != self.nopt - 1:
                    starting_point = stats.reciprocal.rvs([j[0] for j in self.bounds], [j[1] for j in self.bounds], 1,
                                                          random_state=self.random_state)
            if min(fun_value) == np.inf:
                raise NotImplementedError("Maximum likelihood estimator failed: Choose different starting point or "
                                          "increase nopt")
            t = np.argmin(fun_value)
            self.corr_model_params = minimizer[t, :]

        # Updated Correlation matrix corresponding to MLE estimates of hyperparameters
        self.R = self.corr_model(x=s_, s=s_, params=self.corr_model_params)
        # Compute the regression coefficient (solving this linear equation: F * beta = Y)
        c = np.linalg.cholesky(self.R)  # Eq: 3.8, DACE
        c_inv = np.linalg.inv(c)
        f_dash = np.linalg.solve(c, self.F)
        y_dash = np.linalg.solve(c, y_)
        q_, g_ = np.linalg.qr(f_dash)  # Eq: 3.11, DACE
        # Check if F is a full rank matrix
        if np.linalg.matrix_rank(g_) != min(np.size(self.F, 0), np.size(self.F, 1)):
            raise NotImplementedError("Chosen regression functions are not sufficiently linearly independent")
        # Design parameters (beta: regression coefficient)
        self.beta = np.linalg.solve(g_, np.matmul(np.transpose(q_), y_dash))

        # Design parameter (R * gamma = Y - F * beta = residual)
        self.gamma = np.linalg.solve(c.T, (y_dash - np.matmul(f_dash, self.beta)))

        # Computing the process variance (Eq: 3.13, DACE)
        self.err_var = np.zeros(output_dim)
        for i in range(output_dim):
            self.err_var[i] = (1 / nsamples) * (np.linalg.norm(y_dash[:, i] - np.matmul(f_dash, self.beta[:, i])) ** 2)

        self.F_dash, self.C_inv, self.G = f_dash, c_inv, g_

        if self.verbose:
            print('UQpy: Kriging fit complete.')

    def predict(self, x, return_std=False):
        """
        Predict the model response at new points.

        This method evaluates the regression and correlation model at new sample points. Then, it predicts the function
        value and standard deviation.

        **Inputs:**

        * **x** (`list` or `numpy array`):
            Points at which to predict the model response.

        * **return_std** (`Boolean`):
            Indicator to estimate standard deviation.

        **Outputs:**

        * **f_x** (`numpy array`):
            Predicted values at the new points.

        * **std_f_x** (`numpy array`):
            Standard deviation of predicted values at the new points.

        """
        x_ = np.atleast_2d(x)
        if self.normalize:
            x_ = (x_ - self.sample_mean) / self.sample_std
            s_ = (self.samples - self.sample_mean) / self.sample_std
        else:
            s_ = self.samples
        fx, jf = self.reg_model(x_)
        rx = self.corr_model(x=x_, s=s_, params=self.corr_model_params)
        y = np.einsum('ij,jk->ik', fx, self.beta) + np.einsum('ij,jk->ik', rx, self.gamma)
        if self.normalize:
            y = self.value_mean + y * self.value_std
        if x_.shape[1] == 1:
            y = y.flatten()
        if return_std:
            r_dash = np.matmul(self.C_inv, rx.T)
            u = np.matmul(self.F_dash.T, r_dash) - fx.T
            norm1 = np.linalg.norm(r_dash, 2, 0)
            norm2 = np.linalg.norm(np.linalg.solve(self.G, u), 2, 0)
            mse = self.err_var * np.atleast_2d(1 + norm2 - norm1).T
            if self.normalize:
                mse = self.value_std * np.sqrt(mse)
            if x_.shape[1] == 1:
                mse = mse.flatten()
            return y, mse
        else:
            return y

    def jacobian(self, x):
        """
        Predict the gradient of the model at new points.

        This method evaluates the regression and correlation model at new sample point. Then, it predicts the gradient
        using the regression coefficients and the training data.

        **Input:**

        * **x** (`list` or `numpy array`):
            Points at which to evaluate the gradient.

        **Output:**

        * **grad_x** (`list` or `numpy array`):
            Gradient of the surrogate model evaluated at the new points.

        """
        x_ = np.atleast_2d(x)
        if self.normalize:
            x_ = (x_ - self.sample_mean) / self.sample_std
            s_ = (self.samples - self.sample_mean) / self.sample_std
        else:
            s_ = self.samples

        fx, jf = self.reg_model(x_)
        rx, drdx = self.corr_model(x=x_, s=s_, params=self.corr_model_params, dx=True)
        y_grad = np.einsum('ikj,jm->ik', jf, self.beta) + np.einsum('ijk,jm->ki', drdx.T, self.gamma)
        if self.normalize:
            y_grad = y_grad * self.value_std / self.sample_std
        if x_.shape[1] == 1:
            y_grad = y_grad.flatten()
        return y_grad

    # Defining Regression model (Linear)
    def _regress(self):
        if self.reg_model == 'Constant':
            def r(s):
                s = np.atleast_2d(s)
                fx = np.ones([np.size(s, 0), 1])
                jf = np.zeros([np.size(s, 0), np.size(s, 1), 1])
                return fx, jf
        elif self.reg_model == 'Linear':
            def r(s):
                s = np.atleast_2d(s)
                fx = np.concatenate((np.ones([np.size(s, 0), 1]), s), 1)
                jf_b = np.zeros([np.size(s, 0), np.size(s, 1), np.size(s, 1)])
                np.einsum('jii->ji', jf_b)[:] = 1
                jf = np.concatenate((np.zeros([np.size(s, 0), np.size(s, 1), 1]), jf_b), 2)
                return fx, jf
        else:
            def r(s):
                s = np.atleast_2d(s)
                fx = np.zeros([np.size(s, 0), int((np.size(s, 1) + 1) * (np.size(s, 1) + 2) / 2)])
                jf = np.zeros(
                    [np.size(s, 0), np.size(s, 1), int((np.size(s, 1) + 1) * (np.size(s, 1) + 2) / 2)])
                for i in range(np.size(s, 0)):
                    temp = np.hstack((1, s[i, :]))
                    for j in range(np.size(s, 1)):
                        temp = np.hstack((temp, s[i, j] * s[i, j::]))
                    fx[i, :] = temp
                    # definie H matrix
                    h_ = 0
                    for j in range(np.size(s, 1)):
                        tmp_ = s[i, j] * np.eye(np.size(s, 1))
                        t1 = np.zeros([np.size(s, 1), np.size(s, 1)])
                        t1[j, :] = s[i, :]
                        tmp = tmp_ + t1
                        if j == 0:
                            h_ = tmp[:, j::]
                        else:
                            h_ = np.hstack((h_, tmp[:, j::]))
                    jf[i, :, :] = np.hstack((np.zeros([np.size(s, 1), 1]), np.eye(np.size(s, 1)), h_))
                return fx, jf

        return r

    # Defining Correlation model (Gaussian Process)
    def _corr(self):
        def check_samples_and_return_stack(x, s):
            x_, s_ = np.atleast_2d(x), np.atleast_2d(s)
            # Create stack matrix, where each block is x_i with all s
            stack = np.tile(np.swapaxes(np.atleast_3d(x_), 1, 2), (1, np.size(s_, 0), 1)) - np.tile(s_, (
                np.size(x_, 0),
                1, 1))
            return stack

        def derivatives(x_, s_, params):
            stack = check_samples_and_return_stack(x_, s_)
            # Taking stack and creating array of all thetaj*dij
            after_parameters = params * abs(stack)
            # Create matrix of all ones to compare
            comp_ones = np.ones((np.size(x_, 0), np.size(s_, 0), np.size(s_, 1)))
            # zeta_matrix has all values min{1,theta*dij}
            zeta_matrix_ = np.minimum(after_parameters, comp_ones)
            # Copy zeta_matrix to another matrix that will used to find where derivative should be zero
            indices = zeta_matrix_.copy()
            # If value of min{1,theta*dij} is 1, the derivative should be 0.
            # So, replace all values of 1 with 0, then perform the .astype(bool).astype(int)
            # operation like in the linear example, so you end up with an array of 1's where
            # the derivative should be caluclated and 0 where it should be zero
            indices[indices == 1] = 0
            # Create matrix of all |dij| (where non zero) to be used in calculation of dR/dtheta
            dtheta_derivs_ = indices.astype(bool).astype(int) * abs(stack)
            # Same as above, but for matrix of all thetaj where non-zero
            dx_derivs_ = indices.astype(bool).astype(int) * params * np.sign(stack)
            return zeta_matrix_, dtheta_derivs_, dx_derivs_

        if self.corr_model == 'Exponential':
            def c(x, s, params, dt=False, dx=False):
                stack = check_samples_and_return_stack(x, s)
                rx = np.exp(np.sum(-params * abs(stack), axis=2))
                if dt:
                    drdt = - abs(stack) * np.transpose(np.tile(rx, (np.size(x, 1), 1, 1)), (1, 2, 0))
                    return rx, drdt
                if dx:
                    drdx = - params * np.sign(stack) * np.transpose(np.tile(rx, (np.size(x, 1), 1, 1)), (1, 2, 0))
                    return rx, drdx
                return rx
        elif self.corr_model == 'Gaussian':
            def c(x, s, params, dt=False, dx=False):
                stack = check_samples_and_return_stack(x, s)
                rx = np.exp(np.sum(-params * (stack ** 2), axis=2))
                if dt:
                    drdt = -(stack ** 2) * np.transpose(np.tile(rx, (np.size(x, 1), 1, 1)), (1, 2, 0))
                    return rx, drdt
                if dx:
                    drdx = - 2 * params * stack * np.transpose(np.tile(rx, (np.size(x, 1), 1, 1)), (1, 2, 0))
                    return rx, drdx
                return rx
        elif self.corr_model == 'Linear':
            def c(x, s, params, dt=False, dx=False):
                stack = check_samples_and_return_stack(x, s)
                # Taking stack and turning each d value into 1-theta*dij
                after_parameters = 1 - params * abs(stack)
                # Define matrix of zeros to compare against (not necessary to be defined separately,
                # but the line is bulky if this isn't defined first, and it is used more than once)
                comp_zero = np.zeros((np.size(x, 0), np.size(s, 0), np.size(s, 1)))
                # Compute matrix of max{0,1-theta*d}
                max_matrix = np.maximum(after_parameters, comp_zero)
                rx = np.prod(max_matrix, 2)
                # Create matrix that has 1s where max_matrix is nonzero
                # -Essentially, this acts as a way to store the indices of where the values are nonzero
                ones_and_zeros = max_matrix.astype(bool).astype(int)
                # Set initial derivatives as if all were positive
                first_dtheta = -abs(stack)
                first_dx = np.negative(params) * np.sign(stack)
                # Multiply derivs by ones_and_zeros...this will set the values where the
                # derivative should be zero to zero, and keep all other values the same
                drdt = np.multiply(first_dtheta, ones_and_zeros)
                drdx = np.multiply(first_dx, ones_and_zeros)
                if dt:
                    # Loop over parameters, shifting max_matrix and multiplying over derivative matrix with each iter
                    for i in range(len(params) - 1):
                        drdt = drdt * np.roll(max_matrix, i + 1, axis=2)
                    return rx, drdt
                if dx:
                    # Loop over parameters, shifting max_matrix and multiplying over derivative matrix with each iter
                    for i in range(len(params) - 1):
                        drdx = drdx * np.roll(max_matrix, i + 1, axis=2)
                    return rx, drdx
                return rx
        elif self.corr_model == 'Spherical':
            def c(x, s, params, dt=False, dx=False):
                zeta_matrix, dtheta_derivs, dx_derivs = derivatives(x_=x, s_=s, params=params)
                # Initial matrices containing derivates for all values in array. Note since
                # dtheta_s and dx_s already accounted for where derivative should be zero, all
                # that must be done is multiplying the |dij| or thetaj matrix on top of a
                # matrix of derivates w.r.t zeta (in this case, dzeta = -1.5+1.5zeta**2)
                drdt = (-1.5 + 1.5 * zeta_matrix ** 2) * dtheta_derivs
                drdx = (-1.5 + 1.5 * zeta_matrix ** 2) * dx_derivs
                # Also, create matrix for values of equation, 1 - 1.5zeta + 0.5zeta**3, for loop
                zeta_function = 1 - 1.5 * zeta_matrix + 0.5 * zeta_matrix ** 3
                rx = np.prod(zeta_function, 2)
                if dt:
                    # Same as previous example, loop over zeta matrix by shifting index
                    for i in range(len(params) - 1):
                        drdt = drdt * np.roll(zeta_function, i + 1, axis=2)
                    return rx, drdt
                if dx:
                    # Same as previous example, loop over zeta matrix by shifting index
                    for i in range(len(params) - 1):
                        drdx = drdx * np.roll(zeta_function, i + 1, axis=2)
                    return rx, drdx
                return rx
        elif self.corr_model == 'Cubic':
            def c(x, s, params, dt=False, dx=False):
                zeta_matrix, dtheta_derivs, dx_derivs = derivatives(x_=x, s_=s, params=params)
                # Initial matrices containing derivates for all values in array. Note since
                # dtheta_s and dx_s already accounted for where derivative should be zero, all
                # that must be done is multiplying the |dij| or thetaj matrix on top of a
                # matrix of derivates w.r.t zeta (in this case, dzeta = -6zeta+6zeta**2)
                drdt = (-6 * zeta_matrix + 6 * zeta_matrix ** 2) * dtheta_derivs
                drdx = (-6 * zeta_matrix + 6 * zeta_matrix ** 2) * dx_derivs
                # Also, create matrix for values of equation, 1 - 3zeta**2 + 2zeta**3, for loop
                zeta_function_cubic = 1 - 3 * zeta_matrix ** 2 + 2 * zeta_matrix ** 3
                rx = np.prod(zeta_function_cubic, 2)
                if dt:
                    # Same as previous example, loop over zeta matrix by shifting index
                    for i in range(len(params) - 1):
                        drdt = drdt * np.roll(zeta_function_cubic, i + 1, axis=2)
                    return rx, drdt
                if dx:
                    # Same as previous example, loop over zeta matrix by shifting index
                    for i in range(len(params) - 1):
                        drdx = drdx * np.roll(zeta_function_cubic, i + 1, axis=2)
                    return rx, drdx
                return rx
        else:
            def c(x, s, params, dt=False, dx=False):
                # x_, s_ = np.atleast_2d(x_), np.atleast_2d(s_)
                # # Create stack matrix, where each block is x_i with all s
                # stack = np.tile(np.swapaxes(np.atleast_3d(x_), 1, 2), (1, np.size(s_, 0), 1)) - np.tile(s_, (
                #     np.size(x_, 0),
                #     1, 1))
                stack = check_samples_and_return_stack(x, s)
                # In this case, the zeta value is just abs(stack)*parameters, no comparison
                zeta_matrix = abs(stack) * params
                # So, dtheta and dx are just |dj| and theta*sgn(dj), respectively
                dtheta_derivs = abs(stack)
                # dx_derivs = np.ones((np.size(x,0),np.size(s,0),np.size(s,1)))*parameters
                dx_derivs = np.sign(stack) * params

                # Initialize empty sigma and dsigma matrices
                sigma = np.ones((zeta_matrix.shape[0], zeta_matrix.shape[1], zeta_matrix.shape[2]))
                dsigma = np.ones((zeta_matrix.shape[0], zeta_matrix.shape[1], zeta_matrix.shape[2]))

                # Loop over cases to create zeta_matrix and subsequent dR matrices
                for i in range(zeta_matrix.shape[0]):
                    for j in range(zeta_matrix.shape[1]):
                        for k in range(zeta_matrix.shape[2]):
                            y = zeta_matrix[i, j, k]
                            if 0 <= y <= 0.2:
                                sigma[i, j, k] = 1 - 15 * y ** 2 + 30 * y ** 3
                                dsigma[i, j, k] = -30 * y + 90 * y ** 2
                            elif 0.2 < y < 1.0:
                                sigma[i, j, k] = 1.25 * (1 - y) ** 3
                                dsigma[i, j, k] = 3.75 * (1 - y) ** 2 * -1
                            elif y >= 1:
                                sigma[i, j, k] = 0
                                dsigma[i, j, k] = 0

                rx = np.prod(sigma, 2)

                if dt:
                    # Initialize derivative matrices incorporating chain rule
                    drdt = dsigma * dtheta_derivs
                    # Loop over to create proper matrices
                    for i in range(len(params) - 1):
                        drdt = drdt * np.roll(sigma, i + 1, axis=2)
                    return rx, drdt
                if dx:
                    # Initialize derivative matrices incorporating chain rule
                    drdx = dsigma * dx_derivs
                    # Loop over to create proper matrices
                    for i in range(len(params) - 1):
                        drdx = drdx * np.roll(sigma, i + 1, axis=2)
                    return rx, drdx
                return rx
        return c


import scipy.stats as stats
from UQpy.Distributions.baseclass import DistributionContinuous1D


class Gamma(DistributionContinuous1D):
    """
    Gamma distribution having probability density function:

    .. math:: f(x|a) = \dfrac{x^{a-1}\exp(-x)}{\Gamma(a)}

    for :math:`x\ge 0`, :math:`a>0`. Here :math:`\Gamma(a)` refers to the Gamma function.

    In this standard form `(loc=0, scale=1)`. Use `loc` and `scale` to shift and scale the distribution. Specifically,
    this is equivalent to computing :math:`f(y)` where :math:`y=(x-loc)/scale`.

    **Inputs:**

    * **a** (`float`):
        shape parameter
    * **loc** (`float`):
        location parameter
    * **scale** (`float`):
        scale parameter

    The following methods are available for ``Gamma``:

    * ``cdf``, ``pdf``, ``log_pdf``, ``icdf``, ``rvs``, ``moments``, ``fit``.
    """
    def __init__(self, a, loc=0., scale=1.):
        super().__init__(a=a, loc=loc, scale=scale, order_params=('a', 'loc', 'scale'))
        self._construct_from_scipy(scipy_name=stats.gamma)



import numpy as np

from UQpy.Inference.InferenceModel import InferenceModel


########################################################################################################################
########################################################################################################################
#                                  Maximum Likelihood Estimation
########################################################################################################################

class MLEstimation:
    """
    Estimate the maximum likelihood parameters of a model given some data.

    **Inputs:**

    * **inference_model** (object of class ``InferenceModel``):
        The inference model that defines the likelihood function.

    * **data** (`ndarray`):
        Available data, `ndarray` of shape consistent with log likelihood function in ``InferenceModel``

    * **optimizer** (callable):
        Optimization algorithm used to compute the mle.

        | This callable takes in as first input the function to be minimized and as second input an initial guess
          (`ndarray` of shape (n_params, )), along with optional keyword arguments if needed, i.e., it is called within
          the code as:
        | `optimizer(func, x0, **kwargs_optimizer)`

        It must return an object with attributes `x` (minimizer) and `fun` (minimum function value).

        Default is `scipy.optimize.minimize`.

    * **kwargs_optimizer**:
        Keyword arguments that will be transferred to the optimizer.

    * **random_state** (None or `int` or ``numpy.random.RandomState`` object):
        Random seed used to initialize the pseudo-random number generator. Default is None.

        If an integer is provided, this sets the seed for an object of ``numpy.random.RandomState``. Otherwise, the
        object itself can be passed directly.

    * **x0** (`ndarray`):
        Starting point(s) for optimization, see `run_estimation`. Default is `None`.

    * **nopt** (`int`):
        Number of iterations that the optimization is run, starting at random initial guesses. See `run_estimation`.
        Default is `None`.

    If both `x0` and `nopt` are `None`, the object is created but the optimization procedure is not run, one must
    call the ``run`` method.

    **Attributes:**

    * **mle** (`ndarray`):
        Value of parameter vector that maximizes the likelihood function.

    * **max_log_like** (`float`):
        Value of the likelihood function at the MLE.

    **Methods:**

    """
    # Authors: Audrey Olivier, Dimitris Giovanis
    # Last Modified: 12/19 by Audrey Olivier

    def __init__(self, inference_model, data, verbose=False, nopt=None, x0=None, optimizer=None, random_state=None,
                 **kwargs_optimizer):

        # Initialize variables
        self.inference_model = inference_model
        if not isinstance(inference_model, InferenceModel):
            raise TypeError('UQpy: Input inference_model should be of type InferenceModel')
        self.data = data
        self.kwargs_optimizer = kwargs_optimizer
        self.random_state = random_state
        if isinstance(self.random_state, int):
            self.random_state = np.random.RandomState(self.random_state)
        elif not isinstance(self.random_state, (type(None), np.random.RandomState)):
            raise TypeError('UQpy: random_state must be None, an int or an np.random.RandomState object.')
        self.verbose = verbose
        if optimizer is None:
            from scipy.optimize import minimize
            self.optimizer = minimize
        elif callable(optimizer):
            self.optimizer = optimizer
        else:
            raise TypeError('UQpy: Input optimizer should be None (set to scipy.optimize.minimize) or a callable.')
        self.mle = None
        self.max_log_like = None
        if self.verbose:
            print('UQpy: Initialization of MLEstimation object completed.')

        # Run the optimization procedure
        if (nopt is not None) or (x0 is not None):
            self.run(nopt=nopt, x0=x0)

    def run(self, nopt=1, x0=None):
        """
        Run the maximum likelihood estimation procedure.

        This function runs the optimization and updates the `mle` and `max_log_like` attributes of the class. When
        learning the parameters of a distribution, if `dist_object` possesses an ``mle`` method this method is used. If
        `x0` or `nopt` are given when creating the ``MLEstimation`` object, this method is called automatically when the
        object is created.

        **Inputs:**

        * **x0** (`ndarray`):
            Initial guess(es) for optimization, `ndarray` of shape `(nstarts, nparams)` or `(nparams, )`, where
            `nstarts` is the number of times the optimizer will be called. Alternatively, the user can provide input
            `nopt` to randomly sample initial guess(es). The identified MLE is the one that yields the maximum log
            likelihood over all calls of the optimizer.

        * **nopt** (`int`):
            Number of iterations that the optimization is run, starting at random initial guesses. It is only used if
            `x0` is not provided. Default is 1.

            The random initial guesses are sampled uniformly between 0 and 1, or uniformly between user-defined bounds
            if an input bounds is provided as a keyword argument to the ``MLEstimation`` object.

        """
        # Run optimization (use x0 if provided, otherwise sample starting point from [0, 1] or bounds)
        if self.verbose:
            print('UQpy: Evaluating maximum likelihood estimate for inference model ' + self.inference_model.name)

        # Case 3: check if the distribution pi has a fit method, can be used for MLE. If not, use optimization below.
        if (self.inference_model.dist_object is not None) and hasattr(self.inference_model.dist_object, 'fit'):
            if not (isinstance(nopt, int) and nopt >= 1):
                raise ValueError('UQpy: nopt should be an integer >= 1.')
            for _ in range(nopt):
                self.inference_model.dist_object.update_params(
                    **{key: None for key in self.inference_model.list_params})
                mle_dict = self.inference_model.dist_object.fit(data=self.data)
                mle_tmp = np.array([mle_dict[key] for key in self.inference_model.list_params])
                max_log_like_tmp = self.inference_model.evaluate_log_likelihood(
                    params=mle_tmp[np.newaxis, :], data=self.data)[0]
                # Save result
                if self.mle is None:
                    self.mle = mle_tmp
                    self.max_log_like = max_log_like_tmp
                elif max_log_like_tmp > self.max_log_like:
                    self.mle = mle_tmp
                    self.max_log_like = max_log_like_tmp

        # Otherwise run optimization
        else:
            if x0 is None:
                if not (isinstance(nopt, int) and nopt >= 1):
                    raise ValueError('UQpy: nopt should be an integer >= 1.')
                from UQpy.Distributions import Uniform
                x0 = Uniform().rvs(
                    nsamples=nopt * self.inference_model.nparams, random_state=self.random_state).reshape(
                    (nopt, self.inference_model.nparams))
                if 'bounds' in self.kwargs_optimizer.keys():
                    bounds = np.array(self.kwargs_optimizer['bounds'])
                    x0 = bounds[:, 0].reshape((1, -1)) + (bounds[:, 1] - bounds[:, 0]).reshape((1, -1)) * x0
            else:
                x0 = np.atleast_2d(x0)
                if x0.shape[1] != self.inference_model.nparams:
                    raise ValueError('UQpy: Wrong dimensions in x0')
            for x0_ in x0:
                res = self.optimizer(self._evaluate_func_to_minimize, x0_, **self.kwargs_optimizer)
                mle_tmp = res.x
                max_log_like_tmp = (-1.) * res.fun
                # Save result
                if self.mle is None:
                    self.mle = mle_tmp
                    self.max_log_like = max_log_like_tmp
                elif max_log_like_tmp > self.max_log_like:
                    self.mle = mle_tmp
                    self.max_log_like = max_log_like_tmp

            if self.verbose:
                print('UQpy: ML estimation completed.')

    def _evaluate_func_to_minimize(self, one_param):
        """
        Compute negative log likelihood for one parameter vector.

        This is the function to be minimized in the optimization procedure. This is a utility function that will not be
        called by the user.

        **Inputs:**

        * **one_param** (`ndarray`):
            A parameter vector, `ndarray` of shape (nparams, ).

        **Output/Returns:**

        * (`float`):
            Value of negative log-likelihood.
        """

        return -1 * self.inference_model.evaluate_log_likelihood(params=one_param.reshape((1, -1)), data=self.data)[0]





import numpy as np
import scipy.stats as stats

from UQpy.SampleMethods.LHS import LHS
########################################################################################################################
########################################################################################################################
#                                  Adaptive Kriging-Monte Carlo Simulation (AK-MCS)
########################################################################################################################

class AKMCS:
    """
    Adaptively sample for construction of a Kriging surrogate for different objectives including reliability,
    optimization, and global fit.


    **Inputs:**

    * **dist_object** ((list of) ``Distribution`` object(s)):
        List of ``Distribution`` objects corresponding to each random variable.

    * **runmodel_object** (``RunModel`` object):
        A ``RunModel`` object, which is used to evaluate the model.

    * **samples** (`ndarray`):
        The initial samples at which to evaluate the model.

        Either `samples` or `nstart` must be provided.

    * **krig_object** (`class` object):
        A Kriging surrogate model, this object must have ``fit`` and ``predict`` methods.

        May be an object of the ``UQpy`` ``Kriging`` class or an object of the ``scikit-learn``
        ``GaussianProcessRegressor``

    * **nsamples** (`int`):
        Total number of samples to be drawn (including the initial samples).

        If `nsamples` and `samples` are provided when instantiating the class, the ``run`` method will automatically be
        called. If either `nsamples` or `samples` is not provided, ``AKMCS`` can be executed by invoking the ``run``
        method and passing `nsamples`.

    * **nlearn** (`int`):
        Number of samples generated for evaluation of the learning function. Samples for the learning set are drawn
        using ``LHS``.

    * **qoi_name** (`dict`):
        Name of the quantity of interest. If the quantity of interest is a dictionary, this is used to convert it to
        a list

    * **learning_function** (`str` or `function`):
        Learning function used as the selection criteria to identify new samples.

        Built-in options:
                    1. 'U' - U-function \n
                    2. 'EFF' - Expected Feasibility Function \n
                    3. 'Weighted-U' - Weighted-U function \n
                    4. 'EIF' - Expected Improvement Function \n
                    5. 'EGIF' - Expected Global Improvement Fit \n

        `learning_function` may also be passed as a user-defined callable function. This function must accept a Kriging
        surrogate model object with ``fit`` and ``predict`` methods, the set of learning points at which to evaluate the
        learning function, and it may also take an arbitrary number of additional parameters that are passed to
        ``AKMCS`` as `**kwargs`.

    * **n_add** (`int`):
            Number of samples to be added per iteration.

            Default: 1.

    * **random_state** (None or `int` or ``numpy.random.RandomState`` object):
        Random seed used to initialize the pseudo-random number generator. Default is None.

        If an integer is provided, this sets the seed for an object of ``numpy.random.RandomState``. Otherwise, the
        object itself can be passed directly.

    * **verbose** (`Boolean`):
        A boolean declaring whether to write text to the terminal.

        Default value: False.

    * **kwargs**
        Used to pass parameters to `learning_function`.

        For built-in `learning_functions`, see the requisite inputs in the method list below.

        For user-defined `learning_functions`, these will be defined by the requisite inputs to the user-defined method.


    **Attributes:**

    * **samples** (`ndarray`):
        `ndarray` containing the samples at which the model is evaluated.

    * **lf_values** (`list`)
        The learning function evaluated at new sample points.


    **Methods:**

    """

    def __init__(self, dist_object, runmodel_object, krig_object, samples=None, nsamples=None, nlearn=None,
                 qoi_name=None, learning_function='U', n_add=1, random_state=None, verbose=False, **kwargs):

        # Initialize the internal variables of the class.
        self.runmodel_object = runmodel_object
        self.samples = np.array(samples)
        self.nlearn = nlearn
        self.nstart = None
        self.verbose = verbose
        self.qoi_name = qoi_name

        self.learning_function = learning_function
        self.learning_set = None
        self.dist_object = dist_object
        self.nsamples = nsamples

        self.moments = None
        self.n_add = n_add
        self.indicator = False
        self.pf = []
        self.cov_pf = []
        self.dimension = 0
        self.qoi = None
        self.krig_model = None
        self.kwargs = kwargs

        # Initialize and run preliminary error checks.
        self.dimension = len(dist_object)

        if samples is not None:
            if self.dimension != self.samples.shape[1]:
                raise NotImplementedError("UQpy Error: Dimension of samples and distribution are inconsistent.")

        if self.learning_function not in ['EFF', 'U', 'Weighted-U', 'EIF', 'EIGF']:
            raise NotImplementedError("UQpy Error: The provided learning function is not recognized.")
        elif self.learning_function == 'EIGF':
            self.learning_function = self.eigf
        elif self.learning_function == 'EIF':
            if 'eif_stop' not in self.kwargs:
                self.kwargs['eif_stop'] = 0.01
            self.learning_function = self.eif
        elif self.learning_function == 'U':
            if 'u_stop' not in self.kwargs:
                self.kwargs['u_stop'] = 2
            self.learning_function = self.u
        elif self.learning_function == 'Weighted-U':
            if 'u_stop' not in self.kwargs:
                self.kwargs['u_stop'] = 2
            self.learning_function = self.weighted_u
        else:
            if 'a' not in self.kwargs:
                self.kwargs['a'] = 0
            if 'epsilon' not in self.kwargs:
                self.kwargs['epsilon'] = 2
            if 'eff_stop' not in self.kwargs:
                self.kwargs['u_stop'] = 0.001
            self.learning_function = self.eff

        from UQpy.Distributions import DistributionContinuous1D, JointInd

        if isinstance(dist_object, list):
            for i in range(len(dist_object)):
                if not isinstance(dist_object[i], DistributionContinuous1D):
                    raise TypeError('UQpy: A DistributionContinuous1D object must be provided.')
        else:
            if not isinstance(dist_object, (DistributionContinuous1D, JointInd)):
                raise TypeError('UQpy: A DistributionContinuous1D or JointInd object must be provided.')

        self.random_state = random_state
        if isinstance(self.random_state, int):
            self.random_state = np.random.RandomState(self.random_state)
        elif not isinstance(self.random_state, (type(None), np.random.RandomState)):
            raise TypeError('UQpy: random_state must be None, an int or an np.random.RandomState object.')

        if hasattr(krig_object, 'fit') and hasattr(krig_object, 'predict'):
            self.krig_object = krig_object
        else:
            raise NotImplementedError("UQpy: krig_object must have 'fit' and 'predict' methods.")

        if self.verbose:
            print('UQpy: AKMCS - Running the initial sample set using RunModel.')

        # Evaluate model at the training points
        if len(self.runmodel_object.qoi_list) == 0 and samples is not None:
            self.runmodel_object.run(samples=self.samples, append_samples=False)
        if samples is not None:
            if len(self.runmodel_object.qoi_list) != self.samples.shape[0]:
                raise NotImplementedError("UQpy: There should be no model evaluation or Number of samples and model "
                                          "evaluation in RunModel object should be same.")

        if self.nsamples is not None:
            if self.nsamples <= 0 or type(self.nsamples).__name__ != 'int':
                raise NotImplementedError("UQpy: Number of samples to be generated 'nsamples' should be a positive "
                                          "integer.")

            if samples is not None:
                self.run(nsamples=self.nsamples)

    def run(self, nsamples, samples=None, append_samples=True, nstart=None):
        """
        Execute the ``AKMCS`` learning iterations.

        The ``run`` method is the function that performs iterations in the ``AKMCS`` class. If `nsamples` is
        provided when defining the ``AKMCS`` object, the ``run`` method is automatically called. The user may also
        call the ``run`` method directly to generate samples. The ``run`` method of the ``AKMCS`` class can be invoked
        many times.

        **Inputs:**

        * **nsamples** (`int`):
            Total number of samples to be drawn (including the initial samples).

        * **samples** (`ndarray`):
            Samples at which to evaluate the model.

        * **nstart** (`int`):
            Number of initial samples, randomly generated using ``LHS`` class.

            Either `samples` or `nstart` must be provided.

        * **append_samples** (`boolean`)
            Append new samples and model evaluations to the existing samples and model evaluations.

            If ``append_samples = False``, all previous samples and the corresponding quantities of interest from their
            model evaluations are deleted.

            If ``append_samples = True``, samples and their resulting quantities of interest are appended to the
            existing ones.

        **Output/Returns:**

        The ``run`` method has no returns, although it creates and/or appends the `samples` attribute of the
        ``AKMCS`` class.

        """

        self.nsamples = nsamples
        self.nstart = nstart

        if samples is not None:
            # New samples are appended to existing samples, if append_samples is TRUE
            if append_samples:
                if len(self.samples.shape) == 0:
                    self.samples = np.array(samples)
                else:
                    self.samples = np.vstack([self.samples, np.array(samples)])
            else:
                self.samples = np.array(samples)
                self.runmodel_object.qoi_list = []

            if self.verbose:
                print('UQpy: AKMCS - Evaluating the model at the sample set using RunModel.')

            self.runmodel_object.run(samples=samples, append_samples=append_samples)
        else:
            if len(self.samples.shape) == 0:
                if self.nstart is None:
                    raise NotImplementedError("UQpy: User should provide either 'samples' or 'nstart' value.")
                if self.verbose:
                    print('UQpy: AKMCS - Generating the initial sample set using Latin hypercube sampling.')
                self.samples = LHS(dist_object=self.dist_object, nsamples=self.nstart,
                                   random_state=self.random_state).samples
                self.runmodel_object.run(samples=self.samples)

        if self.verbose:
            print('UQpy: Performing AK-MCS design...')

        # If the quantity of interest is a dictionary, convert it to a list
        self.qoi = [None] * len(self.runmodel_object.qoi_list)
        if type(self.runmodel_object.qoi_list[0]) is dict:
            for j in range(len(self.runmodel_object.qoi_list)):
                self.qoi[j] = self.runmodel_object.qoi_list[j][self.qoi_name]
        else:
            self.qoi = self.runmodel_object.qoi_list

        # Train the initial Kriging model.
        self.krig_object.fit(self.samples, self.qoi)
        self.krig_model = self.krig_object.predict

        # kwargs = {"n_add": self.n_add, "parameters": self.kwargs, "samples": self.samples, "qoi": self.qoi,
        #           "dist_object": self.dist_object}

        # ---------------------------------------------
        # Primary loop for learning and adding samples.
        # ---------------------------------------------

        for i in range(self.samples.shape[0], self.nsamples):
            # Initialize the population of samples at which to evaluate the learning function and from which to draw
            # in the sampling.

            lhs = LHS(dist_object=self.dist_object, nsamples=self.nlearn, random_state=self.random_state)
            self.learning_set = lhs.samples.copy()

            # Find all of the points in the population that have not already been integrated into the training set
            rest_pop = np.array([x for x in self.learning_set.tolist() if x not in self.samples.tolist()])

            # Apply the learning function to identify the new point to run the model.

            # new_point, lf, ind = self.learning_function(self.krig_model, rest_pop, **kwargs)
            new_point, lf, ind = self.learning_function(self.krig_model, rest_pop, n_add=self.n_add,
                                                        parameters=self.kwargs, samples=self.samples, qoi=self.qoi,
                                                        dist_object=self.dist_object)

            # Add the new points to the training set and to the sample set.
            self.samples = np.vstack([self.samples, np.atleast_2d(new_point)])

            # Run the model at the new points
            self.runmodel_object.run(samples=new_point, append_samples=True)

            # If the quantity of interest is a dictionary, convert it to a list
            self.qoi = [None] * len(self.runmodel_object.qoi_list)
            if type(self.runmodel_object.qoi_list[0]) is dict:
                for j in range(len(self.runmodel_object.qoi_list)):
                    self.qoi[j] = self.runmodel_object.qoi_list[j][self.qoi_name]
            else:
                self.qoi = self.runmodel_object.qoi_list

            # Retrain the surrogate model
            self.krig_object.fit(self.samples, self.qoi, nopt=1)
            self.krig_model = self.krig_object.predict

            # Exit the loop, if error criteria is satisfied
            if ind:
                print("UQpy: Learning stops at iteration: ", i)
                break

            if self.verbose:
                print("Iteration:", i)

        if self.verbose:
            print('UQpy: AKMCS complete')

    # ------------------
    # LEARNING FUNCTIONS
    # ------------------
    @staticmethod
    def eigf(surr, pop, **kwargs):
        """
        Expected Improvement for Global Fit (EIGF) learning function. See [7]_ for a detailed explanation.


        **Inputs:**

        * **surr** (`class` object):
            A Kriging surrogate model, this object must have a ``predict`` method as defined in `krig_object` parameter.

        * **pop** (`ndarray`):
            An array of samples defining the learning set at which points the EIGF is evaluated

        * **n_add** (`int`):
            Number of samples to be added per iteration.

            Default: 1.

        * **parameters** (`dictionary`)
            Dictionary containing all necessary parameters and the stopping criterion for the learning function. For
            ``EIGF``, this dictionary is empty as no stopping criterion is specified.

        * **samples** (`ndarray`):
            The initial samples at which to evaluate the model.

        * **qoi** (`list`):
            A list, which contaains the model evaluations.

        * **dist_object** ((list of) ``Distribution`` object(s)):
            List of ``Distribution`` objects corresponding to each random variable.


        **Output/Returns:**

        * **new_samples** (`ndarray`):
            Samples selected for model evaluation.

        * **indicator** (`boolean`):
            Indicator for stopping criteria.

            `indicator = True` specifies that the stopping criterion has been met and the AKMCS.run method stops.

        * **eigf_lf** (`ndarray`)
            EIGF learning function evaluated at the new sample points.

        """
        samples = kwargs['samples']
        qoi = kwargs['qoi']
        n_add = kwargs['n_add']

        g, sig = surr(pop, True)

        # Remove the inconsistency in the shape of 'g' and 'sig' array
        g = g.reshape([pop.shape[0], 1])
        sig = sig.reshape([pop.shape[0], 1])

        # Evaluation of the learning function
        # First, find the nearest neighbor in the training set for each point in the population.
        from sklearn.neighbors import NearestNeighbors
        knn = NearestNeighbors(n_neighbors=1)
        knn.fit(np.atleast_2d(samples))
        neighbors = knn.kneighbors(np.atleast_2d(pop), return_distance=False)

        # noinspection PyTypeChecker
        qoi_array = np.array([qoi[x] for x in np.squeeze(neighbors)])

        # Compute the learning function at every point in the population.
        u = np.square(g - qoi_array) + np.square(sig)
        rows = u[:, 0].argsort()[(np.size(g) - n_add):]

        indicator = False
        new_samples = pop[rows, :]
        eigf_lf = u[rows, :]

        return new_samples, eigf_lf, indicator

    @staticmethod
    def u(surr, pop, **kwargs):
        """
        U-function for reliability analysis. See [3] for a detailed explanation.


        **Inputs:**

        * **surr** (`class` object):
            A Kriging surrogate model, this object must have a ``predict`` method as defined in `krig_object` parameter.

        * **pop** (`ndarray`):
            An array of samples defining the learning set at which points the U-function is evaluated

        * **n_add** (`int`):
            Number of samples to be added per iteration.

            Default: 1.

        * **parameters** (`dictionary`)
            Dictionary containing all necessary parameters and the stopping criterion for the learning function. Here
            this includes the parameter `u_stop`.

        * **samples** (`ndarray`):
            The initial samples at which to evaluate the model.

        * **qoi** (`list`):
            A list, which contaains the model evaluations.

        * **dist_object** ((list of) ``Distribution`` object(s)):
            List of ``Distribution`` objects corresponding to each random variable.


        **Output/Returns:**

        * **new_samples** (`ndarray`):
            Samples selected for model evaluation.

        * **indicator** (`boolean`):
            Indicator for stopping criteria.

            `indicator = True` specifies that the stopping criterion has been met and the AKMCS.run method stops.

        * **u_lf** (`ndarray`)
            U learning function evaluated at the new sample points.

        """
        n_add = kwargs['n_add']
        parameters = kwargs['parameters']

        g, sig = surr(pop, True)

        # Remove the inconsistency in the shape of 'g' and 'sig' array
        g = g.reshape([pop.shape[0], 1])
        sig = sig.reshape([pop.shape[0], 1])

        u = abs(g) / sig
        rows = u[:, 0].argsort()[:n_add]

        indicator = False
        if min(u[:, 0]) >= parameters['u_stop']:
            indicator = True

        new_samples = pop[rows, :]
        u_lf = u[rows, 0]
        return new_samples, u_lf, indicator

    @staticmethod
    def weighted_u(surr, pop, **kwargs):
        """
        Probability Weighted U-function for reliability analysis. See [5]_ for a detailed explanation.


        **Inputs:**

        * **surr** (`class` object):
            A Kriging surrogate model, this object must have a ``predict`` method as defined in `krig_object` parameter.

        * **pop** (`ndarray`):
            An array of samples defining the learning set at which points the weighted U-function is evaluated

        * **n_add** (`int`):
            Number of samples to be added per iteration.

            Default: 1.

        * **parameters** (`dictionary`)
            Dictionary containing all necessary parameters and the stopping criterion for the learning function. Here
            this includes the parameter `u_stop`.

        * **samples** (`ndarray`):
            The initial samples at which to evaluate the model.

        * **qoi** (`list`):
            A list, which contaains the model evaluations.

        * **dist_object** ((list of) ``Distribution`` object(s)):
            List of ``Distribution`` objects corresponding to each random variable.

        **Output/Returns:**

        * **new_samples** (`ndarray`):
            Samples selected for model evaluation.

        * **w_lf** (`ndarray`)
            Weighted U learning function evaluated at the new sample points.

        * **indicator** (`boolean`):
            Indicator for stopping criteria.

            `indicator = True` specifies that the stopping criterion has been met and the AKMCS.run method stops.

        """
        n_add = kwargs['n_add']
        parameters = kwargs['parameters']
        samples = kwargs['samples']
        dist_object = kwargs['dist_object']

        g, sig = surr(pop, True)

        # Remove the inconsistency in the shape of 'g' and 'sig' array
        g = g.reshape([pop.shape[0], 1])
        sig = sig.reshape([pop.shape[0], 1])

        u = abs(g) / sig
        p1, p2 = np.ones([pop.shape[0], pop.shape[1]]), np.ones([samples.shape[0], pop.shape[1]])
        for j in range(samples.shape[1]):
            p1[:, j] = dist_object[j].pdf(np.atleast_2d(pop[:, j]).T)
            p2[:, j] = dist_object[j].pdf(np.atleast_2d(samples[:, j]).T)

        p1 = p1.prod(1).reshape(u.size, 1)
        max_p = max(p2.prod(1))
        u_ = u * ((max_p - p1) / max_p)
        rows = u_[:, 0].argsort()[:n_add]

        indicator = False
        if min(u[:, 0]) >= parameters['weighted_u_stop']:
            indicator = True

        new_samples = pop[rows, :]
        w_lf = u_[rows, :]
        return new_samples, w_lf, indicator

    @staticmethod
    def eff(surr, pop, **kwargs):
        """
        Expected Feasibility Function (EFF) for reliability analysis, see [6]_ for a detailed explanation.


        **Inputs:**

        * **surr** (`class` object):
            A Kriging surrogate model, this object must have a ``predict`` method as defined in `krig_object` parameter.

        * **pop** (`ndarray`):
            An array of samples defining the learning set at which points the EFF is evaluated

        * **n_add** (`int`):
            Number of samples to be added per iteration.

            Default: 1.

        * **parameters** (`dictionary`)
            Dictionary containing all necessary parameters and the stopping criterion for the learning function. Here
            these include `a`, `epsilon`, and `eff_stop`.

        * **samples** (`ndarray`):
            The initial samples at which to evaluate the model.

        * **qoi** (`list`):
            A list, which contaains the model evaluations.

        * **dist_object** ((list of) ``Distribution`` object(s)):
            List of ``Distribution`` objects corresponding to each random variable.


        **Output/Returns:**

        * **new_samples** (`ndarray`):
            Samples selected for model evaluation.

        * **indicator** (`boolean`):
            Indicator for stopping criteria.

            `indicator = True` specifies that the stopping criterion has been met and the AKMCS.run method stops.

        * **eff_lf** (`ndarray`)
            EFF learning function evaluated at the new sample points.

        """
        n_add = kwargs['n_add']
        parameters = kwargs['parameters']

        g, sig = surr(pop, True)

        # Remove the inconsistency in the shape of 'g' and 'sig' array
        g = g.reshape([pop.shape[0], 1])
        sig = sig.reshape([pop.shape[0], 1])
        # Reliability threshold: a_ = 0
        # EGRA method: epsilon = 2*sigma(x)
        a_, ep = parameters['eff_a'], parameters['eff_epsilon']*sig
        t1 = (a_ - g) / sig
        t2 = (a_ - ep - g) / sig
        t3 = (a_ + ep - g) / sig
        eff = (g - a_) * (2 * stats.norm.cdf(t1) - stats.norm.cdf(t2) - stats.norm.cdf(t3))
        eff += -sig * (2 * stats.norm.pdf(t1) - stats.norm.pdf(t2) - stats.norm.pdf(t3))
        eff += ep * (stats.norm.cdf(t3) - stats.norm.cdf(t2))
        rows = eff[:, 0].argsort()[-n_add:]

        indicator = False
        if max(eff[:, 0]) <= parameters['eff_stop']:
            indicator = True

        new_samples = pop[rows, :]
        eff_lf = eff[rows, :]
        return new_samples, eff_lf, indicator

    @staticmethod
    def eif(surr, pop, **kwargs):
        """
        Expected Improvement Function (EIF) for Efficient Global Optimization (EFO). See [4]_ for a detailed
        explanation.


        **Inputs:**

        * **surr** (`class` object):
            A Kriging surrogate model, this object must have a ``predict`` method as defined in `krig_object` parameter.

        * **pop** (`ndarray`):
            An array of samples defining the learning set at which points the EIF is evaluated

        * **n_add** (`int`):
            Number of samples to be added per iteration.

            Default: 1.

        * **parameters** (`dictionary`)
            Dictionary containing all necessary parameters and the stopping criterion for the learning function. Here
            this includes the parameter `eif_stop`.

        * **samples** (`ndarray`):
            The initial samples at which to evaluate the model.

        * **qoi** (`list`):
            A list, which contaains the model evaluations.

        * **dist_object** ((list of) ``Distribution`` object(s)):
            List of ``Distribution`` objects corresponding to each random variable.


        **Output/Returns:**

        * **new_samples** (`ndarray`):
            Samples selected for model evaluation.

        * **indicator** (`boolean`):
            Indicator for stopping criteria.

            `indicator = True` specifies that the stopping criterion has been met and the AKMCS.run method stops.

        * **eif_lf** (`ndarray`)
            EIF learning function evaluated at the new sample points.
        """
        n_add = kwargs['n_add']
        parameters = kwargs['parameters']
        qoi = kwargs['qoi']

        g, sig = surr(pop, True)

        # Remove the inconsistency in the shape of 'g' and 'sig' array
        g = g.reshape([pop.shape[0], 1])
        sig = sig.reshape([pop.shape[0], 1])

        fm = min(qoi)
        eif = (fm - g) * stats.norm.cdf((fm - g) / sig) + sig * stats.norm.pdf((fm - g) / sig)
        rows = eif[:, 0].argsort()[(np.size(g) - n_add):]

        indicator = False
        if max(eif[:, 0]) / abs(fm) <= parameters['eif_stop']:
            indicator = True

        new_samples = pop[rows, :]
        eif_lf = eif[rows, :]
        return new_samples, eif_lf, indicator


import scipy.stats as stats
from UQpy.Distributions.baseclass import DistributionDiscrete1D


class Binomial(DistributionDiscrete1D):
    """
    Binomial distribution having probability mass function:

    .. math:: f(x) = {n \choose x} p^x(1-p)^{n-x}

    for :math:`x\ in \{0, 1, 2, ..., n\}`.

    In this standard form `(loc=0)`. Use `loc` to shift the distribution. Specifically, this is equivalent to computing
    :math:`f(y)` where :math:`y=x-loc`.

    **Inputs:**

    * **n** (`int`):
        number of trials, integer >= 0
    * **p** (`float`):
        success probability for each trial, real number in [0, 1]
    * **loc** (`float`):
        location parameter

    The following methods are available for ``Binomial``:

    * ``cdf``, ``pmf``, ``log_pmf``, ``icdf``, ``rvs``, ``moments``.
    """
    def __init__(self, n, p, loc=0.):
        super().__init__(n=n, p=p, loc=loc, order_params=('n', 'p', 'loc'))
        self._construct_from_scipy(scipy_name=stats.binom)



import scipy.stats as stats
from UQpy.Distributions.baseclass import DistributionDiscrete1D


class Poisson(DistributionDiscrete1D):
    """
    Poisson distribution having probability mass function:

    .. math:: f(x) = \exp{(-\mu)}\dfrac{\mu^k}{k!}

    for :math:`x\ge 0`.

    In this standard form `(loc=0)`. Use `loc` to shift the distribution. Specifically, this is equivalent to computing
    :math:`f(y)` where :math:`y=x-loc`.

    **Inputs:**

    * **mu** (`float`):
        shape parameter
    * **loc** (`float`):
        location parameter

    The following methods are available for ``Poisson``:

    * ``cdf``, ``pmf``, ``log_pmf``, ``icdf``, ``rvs``, ``moments``.
    """
    def __init__(self, mu, loc=0.):
        super().__init__(mu=mu, loc=loc, order_params=('mu', 'loc'))
        self._construct_from_scipy(scipy_name=stats.poisson)



from UQpy.SampleMethods.Strata.delaunay import DelaunayStrata
from UQpy.SampleMethods.Strata.rectangular import RectangularStrata
from UQpy.SampleMethods.Strata.voronoi import VoronoiStrata
from UQpy.SampleMethods.Strata.strata import Strata





from UQpy.SampleMethods.MCMC.mcmc import MCMC
from UQpy.Distributions import *
import numpy as np

class MMH(MCMC):
    """

    Component-wise Modified Metropolis-Hastings algorithm.

    In this algorithm, candidate samples are drawn separately in each dimension, thus the proposal consists of a list
    of 1d distributions. The target pdf can be given as a joint pdf or a list of marginal pdfs in all dimensions. This
    will trigger two different algorithms.

    **References:**

    1. S.-K. Au and J. L. Beck,“Estimation of small failure probabilities in high dimensions by subset simulation,”
       Probabilistic Eng. Mech., vol. 16, no. 4, pp. 263–277, Oct. 2001.

    **Algorithm-specific inputs:**

    * **proposal** ((`list` of) ``Distribution`` object(s)):
        Proposal distribution(s) in one dimension, must have a log_pdf/pdf and rvs method.

        The proposal object may be a list of ``DistributionContinuous1D`` objects or a ``JointInd`` object.
        Default: standard normal

    * **proposal_is_symmetric** ((`list` of) `bool`):
        Indicates whether the proposal distribution is symmetric, affects computation of acceptance probability alpha
        Default: False, set to True if default proposal is used

    **Methods:**

    """
    def __init__(self, pdf_target=None, log_pdf_target=None, args_target=None, nburn=0, jump=1, dimension=None,
                 seed=None, save_log_pdf=False, concat_chains=True, nsamples=None, nsamples_per_chain=None,
                 proposal=None, proposal_is_symmetric=False, verbose=False, random_state=None, nchains=None):

        super().__init__(pdf_target=pdf_target, log_pdf_target=log_pdf_target, args_target=args_target,
                         dimension=dimension, seed=seed, nburn=nburn, jump=jump, save_log_pdf=save_log_pdf,
                         concat_chains=concat_chains, verbose=verbose, random_state=random_state, nchains=nchains)

        # If proposal is not provided: set it as a list of standard gaussians
        from UQpy.Distributions import Normal
        self.proposal = proposal
        self.proposal_is_symmetric = proposal_is_symmetric

        # set default proposal
        if self.proposal is None:
            self.proposal = [Normal(), ] * self.dimension
            self.proposal_is_symmetric = [True, ] * self.dimension
        # Proposal is provided, check it
        else:
            # only one Distribution is provided, check it and transform it to a list
            if isinstance(self.proposal, JointInd):
                self.proposal = [m for m in self.proposal.marginals]
                if len(self.proposal) != self.dimension:
                    raise ValueError('UQpy: Proposal given as a list should be of length dimension')
                [self._check_methods_proposal(p) for p in self.proposal]
            elif not isinstance(self.proposal, list):
                self._check_methods_proposal(self.proposal)
                self.proposal = [self.proposal] * self.dimension
            else:  # a list of proposals is provided
                if len(self.proposal) != self.dimension:
                    raise ValueError('UQpy: Proposal given as a list should be of length dimension')
                [self._check_methods_proposal(p) for p in self.proposal]

        # check the symmetry of proposal, assign False as default
        if isinstance(self.proposal_is_symmetric, bool):
            self.proposal_is_symmetric = [self.proposal_is_symmetric, ] * self.dimension
        elif not (isinstance(self.proposal_is_symmetric, list) and
                  all(isinstance(b_, bool) for b_ in self.proposal_is_symmetric)):
            raise TypeError('UQpy: Proposal_is_symmetric should be a (list of) boolean(s)')

        # check with algo type is used
        if self.evaluate_log_target_marginals is not None:
            self.target_type = 'marginals'
            self.current_log_pdf_marginals = None
        else:
            self.target_type = 'joint'

        if self.verbose:
            print('\nUQpy: Initialization of ' + self.__class__.__name__ + ' algorithm complete.')

        # If nsamples is provided, run the algorithm
        if (nsamples is not None) or (nsamples_per_chain is not None):
            self.run(nsamples=nsamples, nsamples_per_chain=nsamples_per_chain)

    def run_one_iteration(self, current_state, current_log_pdf):
        """
        Run one iteration of the MCMC chain for MMH algorithm, starting at current state - see ``MCMC`` class.
        """
        # The target pdf is provided via its marginals
        accept_vec = np.zeros((self.nchains, ))
        if self.target_type == 'marginals':
            # Evaluate the current log_pdf
            if self.current_log_pdf_marginals is None:
                self.current_log_pdf_marginals = [self.evaluate_log_target_marginals[j](current_state[:, j, np.newaxis])
                                                  for j in range(self.dimension)]

            # Sample candidate (independently in each dimension)
            for j in range(self.dimension):
                candidate_j = current_state[:, j, np.newaxis] + self.proposal[j].rvs(
                    nsamples=self.nchains, random_state=self.random_state)

                # Compute log_pdf_target of candidate sample
                log_p_candidate_j = self.evaluate_log_target_marginals[j](candidate_j)

                # Compute acceptance ratio
                if self.proposal_is_symmetric[j]:  # proposal is symmetric
                    log_ratios = log_p_candidate_j - self.current_log_pdf_marginals[j]
                else:  # If the proposal is non-symmetric, one needs to account for it in computing acceptance ratio
                    log_prop_j = self.proposal[j].log_pdf
                    log_proposal_ratio = (log_prop_j(candidate_j - current_state[:, j, np.newaxis]) -
                                          log_prop_j(current_state[:, j, np.newaxis] - candidate_j))
                    log_ratios = log_p_candidate_j - self.current_log_pdf_marginals[j] - log_proposal_ratio

                # Compare candidate with current sample and decide or not to keep the candidate
                unif_rvs = Uniform().rvs(nsamples=self.nchains, random_state=self.random_state).reshape((-1,))
                for nc, (cand, log_p_cand, r_) in enumerate(
                        zip(candidate_j, log_p_candidate_j, log_ratios)):
                    accept = np.log(unif_rvs[nc]) < r_
                    if accept:
                        current_state[nc, j] = cand
                        self.current_log_pdf_marginals[j][nc] = log_p_cand
                        current_log_pdf = np.sum(self.current_log_pdf_marginals)
                        accept_vec[nc] += 1. / self.dimension

        # The target pdf is provided as a joint pdf
        else:
            candidate = np.copy(current_state)
            for j in range(self.dimension):
                candidate_j = current_state[:, j, np.newaxis] + self.proposal[j].rvs(
                    nsamples=self.nchains, random_state=self.random_state)
                candidate[:, j] = candidate_j[:, 0]

                # Compute log_pdf_target of candidate sample
                log_p_candidate = self.evaluate_log_target(candidate)

                # Compare candidate with current sample and decide or not to keep the candidate
                if self.proposal_is_symmetric[j]:  # proposal is symmetric
                    log_ratios = log_p_candidate - current_log_pdf
                else:  # If the proposal is non-symmetric, one needs to account for it in computing acceptance ratio
                    log_prop_j = self.proposal[j].log_pdf
                    log_proposal_ratio = (log_prop_j(candidate_j - current_state[:, j, np.newaxis]) -
                                          log_prop_j(current_state[:, j, np.newaxis] - candidate_j))
                    log_ratios = log_p_candidate - current_log_pdf - log_proposal_ratio
                unif_rvs = Uniform().rvs(nsamples=self.nchains, random_state=self.random_state).reshape((-1,))
                for nc, (cand, log_p_cand, r_) in enumerate(zip(candidate_j, log_p_candidate, log_ratios)):
                    accept = np.log(unif_rvs[nc]) < r_
                    if accept:
                        current_state[nc, j] = cand
                        current_log_pdf[nc] = log_p_cand
                        accept_vec[nc] += 1. / self.dimension
                    else:
                        candidate[:, j] = current_state[:, j]
        # Update the acceptance rate
        self._update_acceptance_rate(accept_vec)
        return current_state, current_log_pdf



import numpy as np
from UQpy.DimensionReduction.baseclass import POD

########################################################################################################################
########################################################################################################################
#                            Higher Order Singular Value Decomposition                                                 #
########################################################################################################################
########################################################################################################################


class HOSVD(POD):
    """
    HOSVD child class is used for higher-order singular value decomposition on the input solutions tensor.

    **Inputs:**

    * **input_sol** (`ndarray`) or (`list`):
        Second order tensor or list containing the solution snapshots. Third dimension or length of list corresponds
        to the number of snapshots.

    * **modes** (`int`):
        Number of modes to keep for dataset reconstruction.

    * **reconstr_perc** (`float`):
        Dataset reconstruction percentage

    **Methods:**
    """

    def __init__(self, input_sol, modes=10**10, reconstr_perc=10**10, verbose=False):

        super().__init__(input_sol, verbose)
        self.verbose = verbose
        self.modes = modes
        self.reconstr_perc = reconstr_perc

    def run(self, get_error=False):
        """
        Executes the HOSVD method in the ''HOSVD'' class.

        **Input:**

        * **get_error** (`Boolean`):
            A boolean declaring whether to return the reconstruction error.

        **Output/Returns:**

        * **reconstructed_solutions** (`ndarray`):
            Second order tensor containing the reconstructed solution snapshots in their initial spatial and
            temporal dimensions.

        * **reduced_solutions** (`ndarray`):
            An array containing the reduced solutions snapshots. The array's dimensions depends on the dimensions
            of input second order tensor and not on the input number of modes or reconstruction percentage.

        """

        if type(self.input_sol) == list:
            x, y, z = self.input_sol[0].shape[0], self.input_sol[0].shape[1], len(self.input_sol)
        else:
            x, y, z = self.input_sol.shape[0], self.input_sol.shape[1], self.input_sol.shape[2]

        a1, a2, a3 = POD.unfold(self.input_sol)

        u1, sig_1, v1 = np.linalg.svd(a1, full_matrices=True)
        u2, sig_2, v2 = np.linalg.svd(a2, full_matrices=True)
        u3, sig_3, v3 = np.linalg.svd(a3, full_matrices=True)

        sig_3_ = np.diag(sig_3)
        hold = np.dot(np.linalg.inv(u3), a3)
        kron_ = np.kron(u1, u2)

        s3 = np.array(np.dot(hold, np.linalg.inv(kron_.T)))

        if self.modes <= 0:
            print('Warning: Invalid input, the number of modes must be positive.')
            return [], []

        elif self.reconstr_perc <= 0:
            print('Warning: Invalid input, the reconstruction percentage is defined in the range (0,100].')
            return [], []

        elif self.modes != 10**10 and self.reconstr_perc != 10**10:
            print('Warning: Either a number of modes or a reconstruction percentage must be chosen, not both.')
            return [], []

        elif type(self.modes) != int:
            print('Warning: The number of modes must be an integer.')
            return [], []

        else:

            if self.modes == 10**10:
                error_ = []
                for i in range(0, x):
                    error_.append(np.sqrt(((sig_3[i + 1:]) ** 2).sum()) / np.sqrt((sig_3 ** 2).sum()))
                    if i == x:
                        error_.append(0)
                error = [i * 100 for i in error_]
                error.reverse()
                perc = error.copy()
                percentage = min(perc, key=lambda x: abs(x-self.reconstr_perc))
                self.modes = perc.index(percentage) + 1

            else:
                if self.modes > x:
                    print("Warning: A number of modes greater than the number of temporal dimensions was given.")
                    print("Number of temporal dimensions is {}.".format(x))

            reduced_solutions = np.dot(u3, sig_3_)
            u3hat = np.dot(u3[:, :self.modes], sig_3_[:self.modes, :self.modes])
            s3hat = np.dot(np.linalg.inv(sig_3_[:self.modes, :self.modes]), s3[:self.modes, :])

            b = np.kron(u1, u2)
            c = np.dot(s3hat, b.T)
            d = np.dot(u3hat[:, :], c)

            reconstructed_solutions = np.zeros((x, y, z))
            for i in range(z):
                reconstructed_solutions[0:x, 0:y, i] = d[i, :].reshape((x, y))

            if self.verbose:
                print("UQpy: Successful execution of HOSVD!")

            if self.modes == 10**10:
                if self.verbose:
                    print('Dataset reconstruction: {0:.3%}'.format(percentage / 100))

            else:
                if get_error:
                    error_rec = np.sqrt(((sig_3[self.modes:]) ** 2).sum()) / np.sqrt((sig_3 ** 2).sum())
                    print("Reduced-order reconstruction error: {0:.3%}".format(error_rec))

            return reconstructed_solutions, reduced_solutions